"""Select the t0 time and lcoation for training"""

import logging
from typing import Optional

import numpy as np
import pandas as pd
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.utils import Location
from ocf_datapipes.utils.geospatial import (
    spatial_coord_type,
)

logger = logging.getLogger(__name__)


@functional_datapipe("pick_locs_and_t0s")
class PickLocationsAndT0sIterDataPipe(IterDataPipe):
    """Datapipe to yield location-time pairs from the input data source."""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        return_all: bool = False,
        shuffle: bool = False,
        time_dim_name: Optional[str] = "time_utc",
    ):
        """
        Datapipe to yield location-time pairs from the input data source.

        Args:
            source_datapipe: Datapipe emitting Xarray Dataset
            return_all: Whether to return all t0-location pairs,
                if True, also returns them in structured order
            shuffle: If `return_all` is True this sets whether the pairs are
                shuffled before being returned.
            time_dim_name: time dimension name, defaulted to 'time_utc'
        """
        super().__init__()
        self.source_datapipe = source_datapipe
        self.return_all = return_all
        self.shuffle = shuffle
        self.time_dim_name = time_dim_name

    def _yield_all_iter(self, xr_dataset):
        # Get the spatial coords
        xr_coord_system, xr_x_dim, xr_y_dim = spatial_coord_type(xr_dataset)

        t_index, x_index = np.meshgrid(
            np.arange(len(xr_dataset[self.time_dim_name])),
            np.arange(len(xr_dataset[xr_x_dim])),
        )

        index_pairs = np.stack((t_index.ravel(), x_index.ravel())).T

        if self.shuffle:
            index_pairs = np.random.permutation(index_pairs)

        # Iterate through all locations in dataset
        for t_index, loc_index in index_pairs:
            # Get the location ID
            loc_id = None
            for id_dim_name in ["pv_system_id", "gsp_id", "station_id"]:
                if id_dim_name in xr_dataset.coords.keys():
                    loc_id = int(xr_dataset[id_dim_name][loc_index].values)

            t0 = xr_dataset[self.time_dim_name][t_index].values
            location = Location(
                coordinate_system=xr_coord_system,
                x=xr_dataset[xr_x_dim][loc_index].values,
                y=xr_dataset[xr_y_dim][loc_index].values,
                id=loc_id,
            )

            yield location, t0

    def _yield_random_iter(self, xr_dataset):
        # Get the spatial coords
        xr_coord_system, xr_x_dim, xr_y_dim = spatial_coord_type(xr_dataset)

        while True:
            loc_index = np.random.randint(0, len(xr_dataset[xr_x_dim]))

            # Get the location ID
            loc_id = None
            for id_dim_name in ["pv_system_id", "gsp_id", "station_id"]:
                if id_dim_name in xr_dataset.coords.keys():
                    loc_id = int(xr_dataset[id_dim_name][loc_index].values)

            location = Location(
                coordinate_system=xr_coord_system,
                x=xr_dataset[xr_x_dim][loc_index].values,
                y=xr_dataset[xr_y_dim][loc_index].values,
                id=loc_id,
            )

            t0 = np.random.choice(xr_dataset[self.time_dim_name].values)

            yield location, t0

    def __iter__(self) -> tuple[Location, pd.Timestamp]:
        xr_dataset = next(iter(self.source_datapipe))

        if self.return_all:
            return self._yield_all_iter(xr_dataset)
        else:
            return self._yield_random_iter(xr_dataset)



"""
This file has a few functions that are used to get GSP (Grid Supply Point) information

The info comes from National Grid ESO.

ESO - Electricity System Operator. General information can be found here
- https://data.nationalgrideso.com/system/gis-boundaries-for-gb-grid-supply-points

get_gsp_metadata_from_eso: gets the gsp metadata
get_gsp_shape_from_eso: gets the shape of the gsp regions

Peter Dudfield
2021-09-13
"""

import logging
import os
from typing import Union

import geopandas as gpd
import pandas as pd
import requests

from ocf_datapipes.utils.geospatial import osgb_to_lon_lat
from ocf_datapipes.utils.pvlive import get_list_of_gsp_ids

logger = logging.getLogger(__name__)

# When saving a file, the columns need to be less than 10 characters -
# - https://github.com/geopandas/geopandas/issues/1417
# - https://en.wikipedia.org/wiki/Shapefile#Limitations
rename_save_columns = {
    "centroid_x": "cen_x",
    "centroid_y": "cen_y",
    "centroid_lat": "cen_lat",
    "centroid_lon": "cen_lon",
}
rename_load_columns = {v: k for k, v in rename_save_columns.items()}


def get_gsp_metadata_from_eso(load_local_file: bool = True, save_local_file: bool = False) -> str:
    """
    Get the metadata for the gsp, from ESO.

    Args:
        calculate_centroid: Load the shape file also, and calculate the Centroid
        load_local_file: Load from a local file, not from ESO
        save_local_file: Save to a local file, only need to do this is Data is updated.

    Returns: Path to local file of ESO data

    """
    logger.debug("Getting GSP shape file")

    local_file = f"{os.path.dirname(os.path.realpath(__file__))}/eso_metadata.csv"

    if not os.path.isfile(local_file):
        logger.debug("There is no local file so going to get it from ESO, and save it afterwards")
        load_local_file = False
        save_local_file = True

    if load_local_file:
        logger.debug("loading local file for ESO metadata")
        metadata = pd.read_csv(local_file)
        # rename the columns to full name
        logger.debug("loading local file for ESO metadata:done")
    else:
        # we now get this from pvlive
        metadata = get_list_of_gsp_ids(return_dataframe=True, return_national=False)

        # drop duplicates
        metadata = metadata.drop_duplicates(subset=["gsp_id"])

        # drop any nans in the gsp is column, and drop National row
        metadata = metadata[metadata["gsp_id"].notnull()]
        metadata = metadata[metadata["gsp_id"] > 0]

        # add in region id
        metadata["region_id"] = metadata["gsp_id"]

    if save_local_file:
        # save file
        metadata.to_csv(local_file)

    return local_file


def get_gsp_shape_from_eso(
    join_duplicates: bool = True,
    load_local_file: bool = True,
    save_local_file: bool = False,
    return_filename: bool = True,
) -> Union[str, gpd.GeoDataFrame]:
    """
    Get the the gsp shape file from ESO (or a local file)

    Args:
        join_duplicates: If True, any RegionIDs which have multiple entries, will be joined
            together to give one entry.
        load_local_file: Load from a local file, not from ESO
        save_local_file: Save to a local file, only need to do this is Data is updated.
        return_filename: option to return location of the file, or geo pandas dataframe

    Returns: Path to local file of GSP shape data
    """
    logger.debug("Loading GSP shape file")

    local_file = f"{os.path.dirname(os.path.realpath(__file__))}/gsp_shape"

    if not os.path.isdir(local_file):
        logger.debug("There is no local file so going to get it from ESO, and save it afterwards")
        load_local_file = False
        save_local_file = True

    if load_local_file:
        logger.debug("loading local file for GSP shape data")
        shape_gpd = gpd.read_file(local_file)
        # rename the columns to full name
        shape_gpd.rename(columns=rename_load_columns, inplace=True)
        logger.debug("loading local file for GSP shape data:done")
    else:
        # call ESO website. There is a possibility that this API will be replaced and its unclear
        # if this original API will stay operational.
        url = (
            "https://data.nationalgrideso.com/backend/dataset/2810092e-d4b2-472f-b955-d8bea01f9ec0/"
            "resource/08534dae-5408-4e31-8639-b579c8f1c50b/download/gsp_regions_20220314.geojson"
        )

        with requests.get(url) as response:
            shape_gpd = gpd.read_file(response.text)

            # calculate the centroid before using - to_crs
            shape_gpd["centroid_x"] = shape_gpd["geometry"].centroid.x
            shape_gpd["centroid_y"] = shape_gpd["geometry"].centroid.y
            (
                shape_gpd["centroid_lon"],
                shape_gpd["centroid_lat"],
            ) = osgb_to_lon_lat(x=shape_gpd["centroid_x"], y=shape_gpd["centroid_y"])

            # Decided not project the shape data to WGS84, as we want to keep
            # all 'batch' data the same projection.
            # However when plotting it may be useful to project to WGS84
            # i.e shape_gpd = shape_gpd.to_crs(WGS84_CRS)

            # TODO is this right?
            # latest geo json does not have region id in it, so add this for the moment
            shape_gpd.sort_values("GSPs", inplace=True)
            shape_gpd.reset_index(inplace=True, drop=True)
            shape_gpd["RegionID"] = range(1, len(shape_gpd) + 1)

    if save_local_file:
        # rename the columns to less than 10 characters
        shape_gpd_to_save = shape_gpd.copy()
        shape_gpd_to_save.rename(columns=rename_save_columns, inplace=True)

        # save file
        shape_gpd_to_save.to_file(local_file)

    # sort
    shape_gpd = shape_gpd.sort_values(by=["RegionID"])

    # join duplicates, currently some GSP shapes are in split into two
    if join_duplicates:
        logger.debug("Removing duplicates by joining geometry together")

        shape_gpd_no_duplicates = shape_gpd.drop_duplicates(subset=["GSPs"]).copy()
        duplicated_raw = shape_gpd[shape_gpd["GSPs"].duplicated()]

        for _, duplicate in duplicated_raw.iterrows():
            # find index in data set with no duplicates
            index_other = shape_gpd_no_duplicates[
                shape_gpd_no_duplicates["GSPs"] == duplicate.GSPs
            ].index

            # join geometries together
            new_geometry = shape_gpd_no_duplicates.loc[index_other, "geometry"].union(
                duplicate.geometry
            )
            # set new geometry
            shape_gpd_no_duplicates.loc[index_other, "geometry"] = new_geometry

        shape_gpd = shape_gpd_no_duplicates

        # sort after removing duplicates
        shape_gpd.sort_values("GSPs", inplace=True)
        shape_gpd.reset_index(inplace=True, drop=True)
        shape_gpd["RegionID"] = range(1, len(shape_gpd) + 1)
        shape_gpd.set_index("RegionID", inplace=True)
        # make sure index starts at 1 and goes to 317

    if return_filename:
        return local_file

    return shape_gpd



"""Datapipes to trigonometric date and time to NumpyBatch"""

import numpy as np
from numpy.typing import NDArray
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.batch import BatchKey


def _get_date_time_in_pi(
    dt: NDArray[np.datetime64],
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    day_of_year = (dt - dt.astype("datetime64[Y]")).astype(int)
    minute_of_day = (dt - dt.astype("datetime64[D]")).astype(int)

    # converting into positions on sin-cos circle
    time_in_pi = (2 * np.pi) * (minute_of_day / (24 * 3600))
    date_in_pi = (2 * np.pi) * (day_of_year / (365 * 24 * 3600))

    return date_in_pi, time_in_pi


@functional_datapipe("add_trigonometric_date_time")
class AddTrigonometricDateTimeIterDataPipe(IterDataPipe):
    """Adds the trigonometric encodings of date of year, time of day to the NumpyBatch"""

    def __init__(self, source_datapipe: IterDataPipe, modality_name: str):
        """
        Adds the sine and cosine of time to the NumpyBatch

        Args:
            source_datapipe: Datapipe of NumpyBatch
            modality_name: Modality to add the time for
        """
        self.source_datapipe = source_datapipe
        self.modality_name = modality_name
        assert self.modality_name in [
            "wind",
        ], f"Trigonometric time not implemented for {self.modality_name}"

    def __iter__(self):
        for np_batch in self.source_datapipe:
            time_utc = np_batch[BatchKey.wind_time_utc]

            times: NDArray[np.datetime64] = time_utc.astype("datetime64[s]")

            date_in_pi, time_in_pi = _get_date_time_in_pi(times)

            # Store
            date_sin_batch_key = BatchKey[self.modality_name + "_date_sin"]
            date_cos_batch_key = BatchKey[self.modality_name + "_date_cos"]
            time_sin_batch_key = BatchKey[self.modality_name + "_time_sin"]
            time_cos_batch_key = BatchKey[self.modality_name + "_time_cos"]

            np_batch[date_sin_batch_key] = np.sin(date_in_pi)
            np_batch[date_cos_batch_key] = np.cos(date_in_pi)
            np_batch[time_sin_batch_key] = np.sin(time_in_pi)
            np_batch[time_cos_batch_key] = np.cos(time_in_pi)

            yield np_batch



"""Convert GSP to Numpy Batch"""

import logging

from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.batch import BatchKey, NumpyBatch
from ocf_datapipes.utils.utils import datetime64_to_float

logger = logging.getLogger(__name__)


def convert_gsp_to_numpy_batch(xr_data):
    """Convert from Xarray to NumpyBatch"""

    example: NumpyBatch = {
        BatchKey.gsp: xr_data.values,
        BatchKey.gsp_t0_idx: xr_data.attrs["t0_idx"],
        BatchKey.gsp_id: xr_data.gsp_id.values,
        BatchKey.gsp_nominal_capacity_mwp: xr_data.isel(time_utc=0)["nominal_capacity_mwp"].values,
        BatchKey.gsp_effective_capacity_mwp: (
            xr_data.isel(time_utc=0)["effective_capacity_mwp"].values
        ),
        BatchKey.gsp_time_utc: datetime64_to_float(xr_data["time_utc"].values),
    }

    # Coordinates
    for batch_key, dataset_key in (
        (BatchKey.gsp_y_osgb, "y_osgb"),
        (BatchKey.gsp_x_osgb, "x_osgb"),
    ):
        if dataset_key in xr_data.coords.keys():
            example[batch_key] = xr_data[dataset_key].values

    return example


@functional_datapipe("convert_gsp_to_numpy_batch")
class ConvertGSPToNumpyBatchIterDataPipe(IterDataPipe):
    """Convert GSP Xarray to NumpyBatch"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Convert GSP Xarray to NumpyBatch object

        Args:
            source_datapipe: Datapipe emitting GSP Xarray object
        """
        super().__init__()
        self.source_datapipe = source_datapipe

    def __iter__(self) -> NumpyBatch:
        """Convert from Xarray to NumpyBatch"""
        for xr_data in self.source_datapipe:
            yield convert_gsp_to_numpy_batch(xr_data)



"""Select PV systems based off their capacity"""

from typing import Union

import numpy as np
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe


@functional_datapipe("filter_pv_systems_on_capacity")
class FilterPVSystemsOnCapacityIterDataPipe(IterDataPipe):
    """Select PV systems based off their capacity"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        min_capacity_watts: Union[int, float] = 0.0,
        max_capacity_watts: Union[int, float] = np.inf,
    ):
        """
        Filter PV systems based off their capacity

        Args:
            source_datapipe: Datapipe of PV data
            min_capacity_watts: Threshold of PV system power in watts. Systems with capacity lower
                than this are dropped.
            max_capacity_watts: Threshold of PV system power in watts. Systems with capacity higher
                than this are dropped.
        """
        self.source_datapipe = source_datapipe
        self.min_capacity_watts = min_capacity_watts
        self.max_capaciity_watts = max_capacity_watts

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for ds in self.source_datapipe:
            too_low = ds.observed_capacity_wp < self.min_capacity_watts
            too_high = ds.observed_capacity_wp > self.max_capacity_watts
            mask = np.logical_or(too_low, too_high)
            yield ds.where(~mask, drop=True)



"""Excarta Loading"""

import numpy as np
import pandas as pd
import xarray as xr

from ocf_datapipes.load.nwp.providers.utils import open_zarr_paths


def preprocess_excarta(ds: xr.Dataset) -> xr.Dataset:
    """
    Preprocesses the Excarta hindcast data

    Args:
        ds: The dataset to preprocess

    Returns:
        The preprocessed dataset, with the init_time_utc added
    """
    # Filename
    filename = ds.encoding["source"]
    # Get the init time from the filename
    init_time = pd.to_datetime(filename.split("/")[-1].split(".")[0])
    # Set the init time
    ds["init_time_utc"] = init_time
    return ds


def open_excarta(zarr_path) -> xr.DataArray:
    """
    Opens the Excarta hindcast data

    Args:
        zarr_path: Path to the zarr to open

    Returns:
        Xarray DataArray of the NWP data
    """

    if "hindcast.zarr" in str(zarr_path):  # Preprocessed one
        nwp = open_zarr_paths(zarr_path)
        nwp = nwp.rename({"__xarray_dataarray_variable__": "excarta"})
        nwp: xr.DataArray = nwp["excarta"]
        time = pd.DatetimeIndex(nwp.init_time_utc)
        assert time.is_unique
        assert time.is_monotonic_increasing
        return nwp
    # Doing the wildcard doesn't really work for Excarta at the moment
    if "*" in str(zarr_path):
        zarr_path = []
        for issue_date in pd.date_range(start="2021-01-01", end="2023-12-31", freq="D"):
            zarr_path.append(
                issue_date.strftime(
                    "https://storage.googleapis.com/excarta-public-us/hindcast/20220225/%Y/%Y%m%d.zarr"
                )
            )
    # Open the data
    nwp: xr.Dataset = open_zarr_paths(
        zarr_path, time_dim="init_time_utc", preprocessor=preprocess_excarta
    )
    nwp = nwp.rename({"prediction_timedelta": "step"})
    nwp = nwp.sortby("init_time_utc")
    # wind is split into speed and direction, so would want to decompose it with sin and cos
    # And split into u and v
    nwp["10u"] = nwp["10m_wind_speed"] * np.cos(np.deg2rad(nwp["10m_wind_speed_angle"]))
    nwp["10v"] = nwp["10m_wind_speed"] * np.sin(np.deg2rad(nwp["10m_wind_speed_angle"]))
    nwp["100u"] = nwp["100m_wind_speed"] * np.cos(np.deg2rad(nwp["100m_wind_speed_angle"]))
    nwp["100v"] = nwp["100m_wind_speed"] * np.sin(np.deg2rad(nwp["100m_wind_speed_angle"]))

    # Combine them all into a channels dimension
    nwp: xr.Dataset = nwp.drop_vars(
        ["10m_wind_speed", "10m_wind_speed_angle", "100m_wind_speed", "100m_wind_speed_angle"]
    )
    nwp: xr.DataArray = nwp.to_array(dim="channel")
    # nwp = nwp.rename({"__xarray_dataarray_variable__": "excarta"})
    # nwp: xr.DataArray = nwp["excarta"]
    # Sanity checks.
    time = pd.DatetimeIndex(nwp.init_time_utc)
    assert time.is_unique
    assert time.is_monotonic_increasing
    return nwp



"""MERRA2 provider loaders"""

import pandas as pd
import xarray as xr

from ocf_datapipes.load.nwp.providers.utils import open_zarr_paths


def open_merra2(zarr_path) -> xr.DataArray:
    """
    Opens the MERRA2 AOD data

    Args:
        zarr_path: Path to the zarr to open

    Returns:
        Xarray DataArray of the NWP data
    """
    # Open the data
    nwp = open_zarr_paths(zarr_path)

    init_time = nwp.time[0]
    nwp = nwp.expand_dims({"init_time_utc": [init_time.values]})
    nwp = nwp.rename({"lat": "latitude", "lon": "longitude", "time": "step"})
    nwp["step"] = nwp["step"] - init_time.values
    nwp = nwp.expand_dims({"channel": list(nwp.keys())}).assign_coords(
        {"channel": list(nwp.keys())}
    )
    nwp = nwp.transpose("init_time_utc", "step", "channel", "latitude", "longitude")
    aodana: xr.DataArray = nwp["AODANA"]
    del nwp

    # Sanity checks.
    time = pd.DatetimeIndex(aodana.step + aodana.init_time_utc.values)
    assert time.is_unique
    assert time.is_monotonic_increasing
    return aodana



"""Convert coordinates Datapipes"""

from typing import Union

import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.utils.geospatial import (
    geostationary_area_coords_to_lonlat,
    lon_lat_to_osgb,
    osgb_to_lon_lat,
)


@functional_datapipe("convert_lonlat_to_osgb")
class ConvertLonLatToOSGBIterDataPipe(IterDataPipe):
    """Convert from Lon/Lat object to OSGB"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Convert from Lon/Lat to OSGB

        Args:
            source_datapipe: Datapipe emitting Xarray objects with latitude and longitude data
        """
        self.source_datapipe = source_datapipe

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """Convert from Lon/Lat to OSGB"""
        for xr_data in self.source_datapipe:
            xr_data["x_osgb"], xr_data["y_osgb"] = lon_lat_to_osgb(
                longitude=xr_data["longitude"],
                latitude=xr_data["latitude"],
            )
            yield xr_data


@functional_datapipe("convert_osgb_to_lonlat")
class ConvertOSGBToLonLatIterDataPipe(IterDataPipe):
    """Convert from OSGB to Lon/Lat"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Convert from OSGB to lon-lat coordinates

        Args:
            source_datapipe: Datapipe emitting Xarray objects with OSGB data
        """
        self.source_datapipe = source_datapipe

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """Convert and add lat/lon to Xarray object"""
        for xr_data in self.source_datapipe:
            xr_data["longitude"], xr_data["latitude"] = osgb_to_lon_lat(
                x=xr_data["x_osgb"], y=xr_data["y_osgb"]
            )
            yield xr_data


@functional_datapipe("convert_geostationary_to_lonlat")
class ConvertGeostationaryToLonLatIterDataPipe(IterDataPipe):
    """Convert from geostationary to Lon/Lat points"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Convert from Geostationary to Lon/Lat points and add to Xarray object

        Args:
            source_datapipe: Datapipe emitting Xarray object with geostationary points
        """
        self.source_datapipe = source_datapipe

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """Convert from geostationary to Lon/Lat and yield the Xarray object"""
        for xr_data in self.source_datapipe:
            xr_data["longitude"], xr_data["latitude"] = geostationary_area_coords_to_lonlat(
                x=xr_data["x_geostationary"],
                y=xr_data["y_geostationary"],
                xr_data=xr_data,
            )

            yield xr_data



"""Create the training/validation datapipe for training the PVNet Model"""

import logging
from datetime import datetime, timedelta
from functools import partial
from typing import List, Optional

import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe
from torch.utils.data.datapipes.iter import IterableWrapper

from ocf_datapipes.batch import MergeNumpyModalities, MergeNWPNumpyModalities
from ocf_datapipes.training.common import (
    DatapipeKeyForker,
    _get_datapipes_dict,
    concat_xr_time_utc,
    construct_loctime_pipelines,
    fill_nans_in_arrays,
    fill_nans_in_pv,
    normalize_gsp,
    normalize_wind,
    potentially_coarsen,
    slice_datapipes_by_time,
)
from ocf_datapipes.utils.consts import (
    METEOMATICS_MEAN,
    METEOMATICS_STDDEV,
    NWP_MEANS,
    NWP_STDS,
    RSS_MEAN,
    RSS_STD,
)
from ocf_datapipes.utils.utils import (
    combine_to_single_dataset,
    flatten_nwp_source_dict,
    nest_nwp_source_dict,
    uncombine_from_single_dataset,
)

xr.set_options(keep_attrs=True)
logger = logging.getLogger("windnet_datapipe")

normalization_values = {
    2019: 3132.0,
    2020: 2817.0,
    2021: 3254.0,
    2022: 3381.0,
    2023: 3225.0,
    2024: 3225.0,
}


def _normalize_wind_power(x: xr.DataArray):
    """Normalize PV data"""
    return x / normalization_values[2024]
    # This is after the data has been temporally sliced, so have the year
    year = x.time_utc.dt.year

    # Add the effective_capacity_mwp to the dataset, indexed on the time_utc
    return (
        x / normalization_values[year]
        if year in normalization_values
        else x / normalization_values[2024]
    )


@functional_datapipe("dict_datasets")
class DictDatasetIterDataPipe(IterDataPipe):
    """Create a dictionary of xr.Datasets from a dict of datapipes"""

    datapipes_dict: dict[IterDataPipe]
    length: Optional[int]

    def __init__(self, datapipes_dict: dict[IterDataPipe]):
        """Init"""
        # Flatten the dict of input datapipes (NWP is nested)
        self.datapipes_dict = flatten_nwp_source_dict(datapipes_dict)
        self.length = None

        # Run checks
        is_okay = all([isinstance(dp, IterDataPipe) for k, dp in self.datapipes_dict.items()])

        if not is_okay:
            raise TypeError(
                "All inputs are required to be `IterDataPipe` " "for `ZipIterDataPipe`."
            )

        super().__init__()

    def __iter__(self):
        """Iter"""
        all_keys = []
        all_datapipes = []
        for k, dp in self.datapipes_dict.items():
            all_keys += [k]
            all_datapipes += [dp]

        zipped_datapipes = all_datapipes[0].zip_ocf(*all_datapipes[1:])

        for values in zipped_datapipes:
            output_dict = {key: x for key, x in zip(all_keys, values)}

            # re-nest the nwp keys
            output_dict = nest_nwp_source_dict(output_dict)

            yield output_dict


@functional_datapipe("load_dict_datasets")
class LoadDictDatasetIterDataPipe(IterDataPipe):
    """Load NetCDF files and split them back into individual xr.Datasets"""

    filenames: List[str]
    keys: List[str]
    nwp_channels: Optional[dict[str, List[str]]] = None
    coarsen_to_deg: Optional[float] = 0.1

    def __init__(
        self,
        filenames: List[str],
        keys: List[str],
        nwp_channels: Optional[dict[str, List[str]]] = None,
        coarsen_to_deg: Optional[float] = 0.1,
    ):
        """
        Load NetCDF files and split them back into individual xr.Datasets

        Args:
            filenames: List of filesnames to load
            keys: List of keys from each file to use, each key should be a
                dataarray in the xr.Dataset
            nwp_channels: Optional dictionary of NWP channels to use
            coarsen_to_deg: what value to coarsen the NWP data to
        """
        super().__init__()
        self.keys = keys
        self.filenames = filenames
        self.nwp_channels = nwp_channels
        self.coarsen_to_deg = coarsen_to_deg

    def __iter__(self):
        """Iterate through each filename, loading it, uncombining it, and then yielding it"""

        while True:
            for filename in self.filenames:
                dataset = xr.open_dataset(filename)
                datasets = uncombine_from_single_dataset(dataset)

                if "ecmwf" in datasets["nwp"].keys():
                    datasets["nwp"]["ecmwf"] = potentially_coarsen(
                        xr_data=datasets["nwp"]["ecmwf"], coarsen_to_deg=self.coarsen_to_deg
                    )

                # Yield a dictionary of the data, using the keys in self.keys
                dataset_dict = {}
                if len(self.keys) > 0:
                    for k in self.keys:
                        try:
                            dataset_dict[k] = datasets[k]
                        except KeyError:
                            continue  # Skip if not available
                else:
                    for k in datasets.keys():
                        dataset_dict[k] = datasets[k]

                # lets reduce the number of nwp channels
                # TODO could move this out to a differnt data pipe, so its more general
                if self.nwp_channels is not None:
                    for key, channels in self.nwp_channels.items():
                        dataset_dict["nwp"][key] = dataset_dict["nwp"][key].sel(channel=channels)

                yield dataset_dict


@functional_datapipe("windnet_convert_to_numpy_batch")
class ConvertToNumpyBatchIterDataPipe(IterDataPipe):
    """Converts Xarray Dataset to Numpy Batch"""

    def __init__(
        self,
        dataset_dict_dp: IterDataPipe,
        check_satellite_no_zeros: bool = False,
    ):
        """Init"""
        super().__init__()
        self.dataset_dict_dp = dataset_dict_dp
        self.check_satellite_no_zeros = check_satellite_no_zeros

    def __iter__(self):
        """Iter"""
        for datapipes_dict in self.dataset_dict_dp:
            # Spatially slice, normalize, and convert data to numpy arrays
            numpy_modalities = []

            if "nwp" in datapipes_dict:
                # Combine the NWPs into NumpyBatch
                nwp_numpy_modalities = dict()
                for nwp_key, nwp_datapipe in datapipes_dict["nwp"].items():
                    nwp_numpy_modalities[nwp_key] = nwp_datapipe.convert_nwp_to_numpy_batch()

                nwp_numpy_modalities = MergeNWPNumpyModalities(nwp_numpy_modalities)
                numpy_modalities.append(nwp_numpy_modalities)

            if "sat" in datapipes_dict:
                numpy_modalities.append(datapipes_dict["sat"].convert_satellite_to_numpy_batch())
            if "pv" in datapipes_dict:
                numpy_modalities.append(datapipes_dict["pv"].convert_pv_to_numpy_batch())
            if "gsp" in datapipes_dict:
                numpy_modalities.append(datapipes_dict["gsp"].convert_gsp_to_numpy_batch())
            if "sensor" in datapipes_dict:
                numpy_modalities.append(datapipes_dict["sensor"].convert_sensor_to_numpy_batch())
            if "wind" in datapipes_dict:
                numpy_modalities.append(datapipes_dict["wind"].convert_wind_to_numpy_batch())

            logger.debug("Combine all the data sources")
            logger.debug("Adding trigonometric date and time")
            combined_datapipe = MergeNumpyModalities(numpy_modalities).add_trigonometric_date_time(
                modality_name="wind"
            )
            # combined_datapipe = MergeNumpyModalities(numpy_modalities).add_sun_position(
            #    modality_name="wind"
            # )

            # logger.info("Filtering out samples with no data")
            # if self.check_satellite_no_zeros:
            # in production we don't want any nans in the satellite data
            #    combined_datapipe = combined_datapipe.map(check_nans_in_satellite_data)

            logger.info("Fill in nans")
            combined_datapipe = combined_datapipe.map(fill_nans_in_arrays)

            yield next(iter(combined_datapipe))


def minutes(num_mins: int):
    """Timedelta of a number of minutes.

    Args:
        num_mins: Minutes timedelta.
    """
    return timedelta(minutes=num_mins)


def construct_sliced_data_pipeline(
    config_filename: str,
    location_pipe: IterDataPipe,
    t0_datapipe: IterDataPipe,
    production: bool = False,
    upsample_nwp: bool = False,
) -> dict:
    """Constructs data pipeline for the input data config file.

    This yields samples from the location and time datapipes.

    Args:
        config_filename: Path to config file.
        location_pipe: Datapipe yielding locations.
        t0_datapipe: Datapipe yielding times.
        production: Whether constucting pipeline for production inference.
        upsample_nwp: Optional to upsample nwp dat Used for ECMWF production data
    """

    datapipes_dict = _get_datapipes_dict(
        config_filename,
        production=production,
    )

    configuration = datapipes_dict.pop("config")

    # Unpack for convenience
    conf_sat = configuration.input_data.satellite
    conf_nwp = configuration.input_data.nwp

    # Slice all of the datasets by time - this is an in-place operation
    slice_datapipes_by_time(datapipes_dict, t0_datapipe, configuration, production)

    # We need a copy of the location datapipe for all keys in fork_keys
    fork_keys = set(k for k in datapipes_dict.keys())
    if "nwp" in datapipes_dict:  # NWP is nested
        fork_keys.update(set(f"nwp/{k}" for k in datapipes_dict["nwp"].keys()))

    # We don't need somes keys even if they are in the data dictionary
    fork_keys = fork_keys - set(
        ["topo", "nwp", "wind", "wind_future", "sensor", "sensor_future", "hrv", "pv_future", "pv"]
    )

    # Set up a key-forker for all the data sources we need it for
    get_loc_datapipe = DatapipeKeyForker(fork_keys, location_pipe)

    if "nwp" in datapipes_dict:
        nwp_datapipes_dict = dict()

        for nwp_key, nwp_datapipe in datapipes_dict["nwp"].items():
            location_pipe, location_pipe_copy = location_pipe.fork(2, buffer_size=5)
            nwp_datapipe = nwp_datapipe.select_spatial_slice_pixels(
                get_loc_datapipe(f"nwp/{nwp_key}"),
                roi_height_pixels=conf_nwp[nwp_key].nwp_image_size_pixels_height,
                roi_width_pixels=conf_nwp[nwp_key].nwp_image_size_pixels_width,
            )
            potentially_coarsen_partial = partial(
                potentially_coarsen, coarsen_to_deg=conf_nwp[nwp_key].coarsen_to_degrees
            )
            nwp_datapipe = nwp_datapipe.map(potentially_coarsen_partial)
            # Somewhat hacky way for India specifically, need different mean/std for ECMWF data
            if conf_nwp[nwp_key].nwp_provider in ["ecmwf"]:
                normalize_provider = "ecmwf_india"
            else:
                normalize_provider = conf_nwp[nwp_key].nwp_provider
            nwp_datapipes_dict[nwp_key] = nwp_datapipe.normalize(
                mean=NWP_MEANS[normalize_provider],
                std=NWP_STDS[normalize_provider],
            )

            if upsample_nwp:
                nwp_datapipes_dict[nwp_key] = nwp_datapipes_dict[nwp_key].upsample(
                    y_upsample=2, x_upsample=2, keep_same_shape=True, round_to_dp=2
                )
    if "sensor" in datapipes_dict:
        sensor_datapipe = datapipes_dict["sensor"]
        sensor_datapipe = sensor_datapipe.select_spatial_slice_meters(
            location_datapipe=get_loc_datapipe("sensor"),
            roi_height_meters=10000000,
            roi_width_meters=100000000,
            dim_name="station_id",
        )
        sensor_datapipe = sensor_datapipe.normalize(mean=METEOMATICS_MEAN, std=METEOMATICS_STDDEV)
    if "sat" in datapipes_dict:
        sat_datapipe = datapipes_dict["sat"]

        sat_datapipe = sat_datapipe.select_spatial_slice_pixels(
            get_loc_datapipe("sat"),
            roi_height_pixels=conf_sat.satellite_image_size_pixels_height,
            roi_width_pixels=conf_sat.satellite_image_size_pixels_width,
        )
        sat_datapipe = sat_datapipe.normalize(mean=RSS_MEAN, std=RSS_STD)

    if "wind" in datapipes_dict:
        # Recombine Sensor arrays - see function doc for further explanation
        wind_datapipe = (
            datapipes_dict["wind"].zip_ocf(datapipes_dict["wind_future"]).map(concat_xr_time_utc)
        )
        wind_datapipe = wind_datapipe.normalize(normalize_fn=normalize_wind)
        wind_datapipe = wind_datapipe.map(fill_nans_in_pv)

    finished_dataset_dict = {"config": configuration}

    if "gsp" in datapipes_dict:
        gsp_future_datapipe = datapipes_dict["gsp_future"]
        gsp_future_datapipe = gsp_future_datapipe.select_spatial_slice_meters(
            location_datapipe=get_loc_datapipe("gsp_future"),
            roi_height_meters=1,
            roi_width_meters=1,
            dim_name="gsp_id",
        )

        gsp_datapipe = datapipes_dict["gsp"]
        gsp_datapipe = gsp_datapipe.select_spatial_slice_meters(
            location_datapipe=get_loc_datapipe("gsp"),
            roi_height_meters=1,
            roi_width_meters=1,
            dim_name="gsp_id",
        )

        # Recombine GSP arrays - see function doc for further explanation
        gsp_datapipe = gsp_datapipe.zip_ocf(gsp_future_datapipe).map(concat_xr_time_utc)
        gsp_datapipe = gsp_datapipe.normalize(normalize_fn=normalize_gsp)

        finished_dataset_dict["gsp"] = gsp_datapipe

    get_loc_datapipe.close()

    if "nwp" in datapipes_dict:
        finished_dataset_dict["nwp"] = nwp_datapipes_dict
    if "sat" in datapipes_dict:
        finished_dataset_dict["sat"] = sat_datapipe
    if "wind" in datapipes_dict:
        finished_dataset_dict["wind"] = wind_datapipe
    if "sensor" in datapipes_dict:
        finished_dataset_dict["sensor"] = sensor_datapipe

    return finished_dataset_dict


def windnet_datapipe(
    config_filename: str,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
) -> IterDataPipe:
    """
    Construct windnet pipeline for the input data config file.

    Args:
        config_filename: Path to config file.
        start_time: Minimum time at which a sample can be selected.
        end_time: Maximum time at which a sample can be selected.
    """
    logger.info("Constructing windnet pipeline")

    # Open datasets from the config and filter to useable location-time pairs
    location_pipe, t0_datapipe = construct_loctime_pipelines(
        config_filename,
        start_time,
        end_time,
    )

    # Shard after we have the loc-times. These are already shuffled so no need to shuffle again
    location_pipe = location_pipe.sharding_filter()
    t0_datapipe = t0_datapipe.sharding_filter()

    # In this function we re-open the datasets to make a clean separation before/after sharding
    # This function
    datapipe_dict = construct_sliced_data_pipeline(
        config_filename,
        location_pipe,
        t0_datapipe,
    )

    # Merge all the datapipes into one
    return DictDatasetIterDataPipe(
        {k: v for k, v in datapipe_dict.items() if k != "config"},
    ).map(combine_to_single_dataset)


def split_dataset_dict_dp(element):
    """
    Wrap each of the data source inputs into a datapipe

    Args:
        element: Dictionary of xarray objects
    """

    element = flatten_nwp_source_dict(element)
    output_dict = {k: IterableWrapper([v]) for k, v in element.items() if k != "config"}
    output_dict = nest_nwp_source_dict(output_dict)

    return output_dict


def windnet_netcdf_datapipe(
    keys: List[str],
    filenames: List[str],
    nwp_channels: Optional[dict[str, List[str]]] = None,
) -> IterDataPipe:
    """
    Load the saved Datapipes from windnet, and transform to numpy batch

    Args:
        config_filename: Path to config file.
        keys: List of keys to extract from the single NetCDF files
        filenames: List of NetCDF files to load
        nwp_channels: Optional dictionary of NWP channels to use

    Returns:
        Datapipe that transforms the NetCDF files to numpy batch
    """
    logger.info("Constructing windnet file pipeline")
    # Load files
    datapipe_dict_dp: IterDataPipe = LoadDictDatasetIterDataPipe(
        filenames=filenames, keys=keys, nwp_channels=nwp_channels
    ).map(split_dataset_dict_dp)
    datapipe = datapipe_dict_dp.windnet_convert_to_numpy_batch()

    return datapipe


if __name__ == "__main__":
    # Load the ECMWF and sensor data here
    # datapipe = windnet_datapipe(
    #    config_filename="/home/jacob/Development/ocf_datapipes/tests/config/india_test.yaml",
    #    start_time=datetime(2023, 1, 1),
    #    end_time=datetime(2023, 11, 2),
    # )
    # batch = next(iter(datapipe))
    # print(batch)
    # batch.to_netcdf("test.nc", engine="h5netcdf")
    # Load the saved NetCDF files here
    datapipe = windnet_netcdf_datapipe(
        keys=["nwp", "wind"],
        filenames=["/run/media/jacob/data/windnet_india_batches_medium/val/000000.nc"],
    )
    batch = next(iter(datapipe))
    # print(batch)



"""Select time periods"""

import datetime
import logging
from typing import Union

import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("filter_times")
class FilterTimesIterDataPipe(IterDataPipe):
    """Select time periods"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        start_time: datetime.datetime,
        end_time: datetime.datetime,
        dim_name: str = "time_utc",
    ):
        """
        Select time periods

        Args:
            source_datapipe: Datapipe of Xarray objects
            start_time: Start time to select
            end_time: End time to select
            dim_name: Dimension name for time
        """
        self.source_datapipe = source_datapipe
        self.start_time = start_time
        self.end_time = end_time
        self.dim_name = dim_name

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for xr_data in self.source_datapipe:
            logger.debug(f"Selecting Train/Test Time period ({self.start_time} - {self.end_time})")
            xr_data = xr_data.sel({self.dim_name: slice(self.start_time, self.end_time)})
            yield xr_data



"""Numpy and xarray batch"""

from enum import Enum, auto
from typing import Union

import numpy as np
import torch
import xarray as xr


class BatchKey(Enum):
    """The names of the different elements of each batch.

    This is also where we document the exact shape of each element.

    Each `DataSource` may be split into several different BatchKey elements. For example, the
    PV PreparedDataSource yields `pv` and `pv_ml_id` BatchKeys.
    """

    # -------------- HRVSATELLITE -----------------------------------
    # shape: (batch_size, time, channels, y, x)
    #
    # Or, if the imagery has been patched,
    # shape: (batch_size, time, channels, y, x, n_pixels_per_patch) where n_pixels_per_patch
    # is the *total* number of pixels,
    # i.e. n_pixels_per_patch_along_height * n_pixels_per_patch_along_width.
    hrvsatellite_actual = auto()
    hrvsatellite_predicted = auto()  # shape: batch_size, time, y, x
    hrvsatellite_t0_idx = auto()  # shape: scalar

    # HRV satellite coordinates:
    hrvsatellite_y_osgb = auto()  # shape: (batch_size, y, x)
    hrvsatellite_x_osgb = auto()  # shape: (batch_size, y, x)
    hrvsatellite_y_geostationary = auto()  # shape: (batch_size, y)
    hrvsatellite_x_geostationary = auto()  # shape: (batch_size, x)
    #: Time is seconds since UNIX epoch (1970-01-01). Shape: (batch_size, n_timesteps)
    hrvsatellite_time_utc = auto()
    # Added by np_batch_processor.Topography:
    hrvsatellite_surface_height = auto()  # The surface height at each pixel. (batch_size, y, x)

    # HRV satellite Fourier coordinates:
    # Spatial coordinates. Shape: (batch_size, y, x, n_fourier_features_per_dim)
    hrvsatellite_y_osgb_fourier = auto()
    hrvsatellite_x_osgb_fourier = auto()
    #: Time shape: (batch_size, n_timesteps, n_fourier_features_per_dim)
    hrvsatellite_time_utc_fourier = auto()

    # -------------- NWP --------------------------------------------
    nwp = auto()  # See `MultiNWPNumpyBatch`

    # -------------- PV ---------------------------------------------
    pv = auto()  # shape: (batch_size, time, n_pv_systems)
    pv_t0_idx = auto()  # shape: scalar
    pv_ml_id = auto()  # shape: (batch_size, n_pv_systems)
    pv_id = auto()  # shape: (batch_size, n_pv_systems)
    pv_observed_capacity_wp = auto()  # shape: (batch_size, n_pv_systems)
    pv_nominal_capacity_wp = auto()  # shape: (batch_size, n_pv_systems)
    #: pv_mask is True for good PV systems in each example.
    # The RawPVDataSource doesn't use pv_mask. Instead is sets missing PV systems to NaN
    # across all PV batch keys.
    pv_mask = auto()  # shape: (batch_size, n_pv_systems)

    # PV coordinates:
    # Each has shape: (batch_size, n_pv_systems), will be NaN for missing PV systems.
    pv_latitude = auto()
    pv_longitude = auto()
    pv_time_utc = auto()  # Seconds since UNIX epoch (1970-01-01).

    # PV Fourier coordinates:
    # Each has shape: (batch_size, n_pv_systems, n_fourier_features_per_dim),
    # and will be NaN for missing PV systems.
    pv_latitude_fourier = auto()
    pv_longitude_fourier = auto()
    pv_time_utc_fourier = auto()  # (batch_size, time, n_fourier_features)

    # -------------- Wind ---------------------------------------------
    wind = auto()  # shape: (batch_size, time, n_pv_systems)
    wind_t0_idx = auto()  # shape: scalar
    wind_ml_id = auto()  # shape: (batch_size, n_pv_systems)
    wind_id = auto()  # shape: (batch_size, n_pv_systems)
    wind_observed_capacity_mwp = auto()  # shape: (batch_size, n_pv_systems)
    wind_nominal_capacity_mwp = auto()  # shape: (batch_size, n_pv_systems)
    #: pv_mask is True for good PV systems in each example.
    # The RawPVDataSource doesn't use pv_mask. Instead is sets missing PV systems to NaN
    # across all PV batch keys.
    wind_mask = auto()  # shape: (batch_size, n_pv_systems)

    # PV coordinates:
    # Each has shape: (batch_size, n_pv_systems), will be NaN for missing PV systems.
    wind_latitude = auto()
    wind_longitude = auto()
    wind_time_utc = auto()  # Seconds since UNIX epoch (1970-01-01).

    # PV Fourier coordinates:
    # Each has shape: (batch_size, n_pv_systems, n_fourier_features_per_dim),
    # and will be NaN for missing PV systems.
    wind_latitude_fourier = auto()
    wind_longitude_fourier = auto()
    wind_time_utc_fourier = auto()  # (batch_size, time, n_fourier_features)

    # -------------- GSP --------------------------------------------
    gsp = auto()  # shape: (batch_size, time, 1)
    gsp_t0_idx = auto()  # shape: scalar
    gsp_id = auto()  # shape: (batch_size)

    # GSP coordinates:
    # Each has shape: (batch_size). No NaNs.
    gsp_y_osgb = auto()
    gsp_x_osgb = auto()
    gsp_time_utc = auto()  # Seconds since UNIX epoch (1970-01-01). (batch_size, time)
    gsp_nominal_capacity_mwp = auto()  # (batch_size)
    gsp_effective_capacity_mwp = auto()  # (batch_size)

    # GSP Fourier coordinates:
    # Each has shape: (batch_size, 1, n_fourier_features_per_dim),
    # no NaNs.
    gsp_y_osgb_fourier = auto()
    gsp_x_osgb_fourier = auto()
    gsp_time_utc_fourier = auto()  # (batch_size, time, n_fourier_features)

    # -------------- TIME -------------------------------------------
    # Sine and cosine of date of year and time of day at every timestep.
    # shape = (batch_size, n_timesteps)
    # This is calculated for wind only inside datapipes.
    wind_date_sin = auto()
    wind_date_cos = auto()
    wind_time_sin = auto()
    wind_time_cos = auto()

    # -------------- SUN --------------------------------------------
    # Solar position at every timestep. shape = (batch_size, n_timesteps)
    # The solar position data comes from two alternative sources: either the Sun pre-prepared
    # batches, or the SunPosition np_batch_processor for the RawDataset.
    hrvsatellite_solar_azimuth = auto()
    hrvsatellite_solar_elevation = auto()
    satellite_solar_azimuth = auto()
    satellite_solar_elevation = auto()
    gsp_solar_azimuth = auto()
    gsp_solar_elevation = auto()
    pv_solar_azimuth = auto()
    pv_solar_elevation = auto()

    # Solar position at the centre of the HRV image at t0
    # (from `power_perceiver.np_batch_processor.SunPosition`)
    # shape = (example,)
    # Not used in Raw data pipeline.
    solar_azimuth_at_t0 = auto()
    solar_elevation_at_t0 = auto()

    # ------------- REQUESTED NUMBER OF TIMESTEPS ---------------------
    # Added by `ReduceNumTimesteps`. Gives the indicies of the randomly selected timesteps.
    # Not used in the Raw data pipeline.
    requested_timesteps = auto()  # shape: (n_requested_timesteps)

    # -------------- SATELLITE -----------------------------------
    # shape: (batch_size, time, channels, y, x)
    #
    # Or, if the imagery has been patched,
    # shape: (batch_size, time, channels, y, x, n_pixels_per_patch) where n_pixels_per_patch
    # is the *total* number of pixels,
    # i.e. n_pixels_per_patch_along_height * n_pixels_per_patch_along_width.
    satellite_actual = auto()
    satellite_predicted = auto()  # shape: batch_size, time, y, x
    satellite_t0_idx = auto()  # shape: scalar

    # HRV satellite coordinates:
    satellite_y_osgb = auto()  # shape: (batch_size, y, x)
    satellite_x_osgb = auto()  # shape: (batch_size, y, x)
    satellite_y_geostationary = auto()  # shape: (batch_size, y)
    satellite_x_geostationary = auto()  # shape: (batch_size, x)
    #: Time is seconds since UNIX epoch (1970-01-01). Shape: (batch_size, n_timesteps)
    satellite_time_utc = auto()
    # Added by np_batch_processor.Topography:
    satellite_surface_height = auto()  # The surface height at each pixel. (batch_size, y, x)

    # HRV satellite Fourier coordinates:
    # Spatial coordinates. Shape: (batch_size, y, x, n_fourier_features_per_dim)
    satellite_y_osgb_fourier = auto()
    satellite_x_osgb_fourier = auto()
    #: Time shape: (batch_size, n_timesteps, n_fourier_features_per_dim)
    satellite_time_utc_fourier = auto()

    # -------------- Sensor ---------------------------------------------
    sensor = auto()  # shape: (batch_size, time, n_pv_systems)
    sensor_t0_idx = auto()  # shape: scalar
    sensor_ml_id = auto()  # shape: (batch_size, n_pv_systems)
    sensor_id = auto()  # shape: (batch_size, n_pv_systems)
    sensor_observed_capacity_wp = auto()  # shape: (batch_size, n_pv_systems)
    sensor_nominal_capacity_wp = auto()  # shape: (batch_size, n_pv_systems)
    #: pv_mask is True for good PV systems in each example.
    # The RawPVDataSource doesn't use pv_mask. Instead is sets missing PV systems to NaN
    # across all PV batch keys.
    sensor_mask = auto()  # shape: (batch_size, n_pv_systems)

    # PV coordinates:
    # Each has shape: (batch_size, n_pv_systems), will be NaN for missing PV systems.
    sensor_latitude = auto()
    sensor_longitude = auto()
    sensor_time_utc = auto()  # Seconds since UNIX epoch (1970-01-01).

    # PV Fourier coordinates:
    # Each has shape: (batch_size, n_pv_systems, n_fourier_features_per_dim),
    # and will be NaN for missing PV systems.
    sensor_latitude_fourier = auto()
    sensor_longitude_fourier = auto()
    sensor_time_utc_fourier = auto()  # (batch_size, time, n_fourier_features)

    wind_solar_azimuth = auto()
    wind_solar_elevation = auto()


class NWPBatchKey(Enum):
    """The names of the different elements of each NWP batch.

    This is also where we document the exact shape of each element.
    """

    # -------------- NWP --------------------------------------------
    nwp = auto()  # shape: (batch_size, target_time_utc, channel, y_osgb, x_osgb)
    nwp_t0_idx = auto()  # shape: scalar
    nwp_target_time_utc = auto()  # shape: (batch_size, target_time_utc)
    nwp_init_time_utc = auto()  # shape: (batch_size, target_time_utc)
    nwp_step = auto()  # Int. Number of hours. shape: (batch_size, target_time_utc)
    nwp_y_osgb = auto()  # shape: (batch_size, y_osgb)
    nwp_x_osgb = auto()  # shape: (batch_size, x_osgb)
    nwp_channel_names = auto()  # shape: (channel,)

    # NWP Fourier features:
    nwp_target_time_utc_fourier = auto()
    nwp_init_time_utc_fourier = auto()
    nwp_y_osgb_fourier = auto()
    nwp_x_osgb_fourier = auto()
    nwp_target_time_solar_azimuth = auto()
    nwp_target_time_solar_elevation = auto()


NWPNumpyBatch = dict[NWPBatchKey, np.ndarray]

NumpyBatch = dict[BatchKey, Union[np.ndarray, dict[str, NWPNumpyBatch]]]

XarrayBatch = dict[BatchKey, Union[xr.DataArray, xr.Dataset]]

TensorBatch = dict[BatchKey, Union[torch.Tensor, dict[str, dict[NWPBatchKey, torch.Tensor]]]]



"""
This is a class function that assigns day night status
"""

import logging

import numpy as np
import pvlib
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)

# {month: [sunrise_hour, sunset_hour]}
uk_daynight_dict = {
    1: [7, 17],
    2: [7, 18],
    3: [9, 19],
    4: [6, 20],
    5: [5, 21],
    6: [4, 22],
    7: [5, 22],
    8: [5, 21],
    9: [6, 20],
    10: [7, 21],
    11: [7, 17],
    12: [7, 17],
}

logger.info(
    f"The day and night standard hours are set by {'https://www.timeanddate.com/sun/uk/london'}"
)


@np.vectorize
def month_to_dawn_hour(month):
    """Function returning the first hour after dawn for the given month.

    Args:
        month: Array-like of ints or int month number in range [1, 12]

    Returns:
        Array of the first hour of the day

    """
    return uk_daynight_dict[month][0]


@np.vectorize
def month_to_dusk_hour(month):
    """Function returning the first hour after dusk for the given month.

    Args:
        month: Array-like of ints or int month number in range [1, 12]

    Returns:
        Array of the first hour of the night

    """
    return uk_daynight_dict[month][1]


@functional_datapipe("assign_daynight_status")
class AssignDayNightStatusIterDataPipe(IterDataPipe):
    """Adds a new dimension of day/night status"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        method: str = "simple",
    ):
        """This method adds extra coordinate of day night status.

        Args:
            source_datapipe: Datapipe emitting Xarray Dataset with time_utc coordinate
            method: Method used to assign the day-night status. Either "simple" or "elevation"

        Notes:
            If method is 'simple', then the month and hour are used to decide which datetimes
            correspond to day and night. These are done via a lookup table compiled for UK-London.
            These will be inappropriate far from London.

            If method is 'elevation', then the sun elevation is calculated for each datetime and PV
            system using its latitude and longitude. This is computationally expensive.
        """

        self.source_datapipe = source_datapipe
        self.method = method

        if method == "simple":
            logger.warning(
                "Calculating the day-night status using method 'simple' is only appropriate for "
                "UK PV datasets"
            )
            self._status_func = self._get_status_by_hour

        elif method == "elevation":
            logger.warning(
                "Calculating the day-night status using method 'elevation' can take a long time"
            )
            self._status_func = self._get_status_by_elevation
        else:
            raise ValueError(f"Method '{method}' not recognised")

    def _get_status_by_hour(self, ds):
        # Getting month and hour values from time_utc and stacking them
        months = ds.time_utc.dt.month.values
        hours = ds.time_utc.dt.hour.values

        # Getting the status of day/night for each timestamp in the dates
        is_after_dawn = hours >= month_to_dawn_hour(months)
        is_before_dusk = hours < month_to_dusk_hour(months)

        status_daynight = np.where(
            np.logical_and(is_after_dawn, is_before_dusk),
            "day",
            "night",
        )

        # Assigning a new coordinates of 'status_daynight' in the DataArray
        ds = ds.assign_coords(status_daynight=(("time_utc"), status_daynight))
        return ds

    def _get_status_by_elevation(self, ds):
        elevation = xr.full_like(ds, fill_value=np.nan).astype(np.float32)

        for system_id in ds.pv_system_id.values[:5]:
            ds_sel = ds.sel(pv_system_id=system_id)

            solpos = pvlib.solarposition.get_solarposition(
                time=ds_sel.time_utc,
                latitude=ds_sel.latitude.item(),
                longitude=ds_sel.longitude.item(),
                # method="nrel_numba",
                # Which `method` to use?
                # pyephem seemed to be a good mix between speed and ease but causes
                # segfaults!
                # nrel_numba doesn't work when using multiple worker processes.
                # nrel_c is probably fastest but requires C code to be manually compiled:
                # https://midcdmz.nrel.gov/spa/
            )
            elevation.sel(pv_system_id=system_id).values[:] = solpos["elevation"]

        status_daynight = np.where(
            elevation > 0.0,
            "day",
            "night",
        )

        # Assigning a new coordinates of 'status_daynight' in the DataArray
        ds = ds.assign_coords(status_daynight=(("time_utc"), status_daynight))
        return ds

    def __iter__(self) -> xr.DataArray():
        """Adds extra coordinate of day-night status

        Returns:
            Dataset with additional coordinate describing the datetime as either 'day' or 'night'
        """

        # Reading the Xarray dataset
        for ds in self.source_datapipe:
            yield self._status_func(ds)



"""Create the training/validation datapipe for training the PVNet Model"""

import logging
from datetime import datetime, timedelta
from typing import List, Optional

import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe
from torch.utils.data.datapipes.iter import IterableWrapper

from ocf_datapipes.batch import BatchKey, MergeNumpyModalities, MergeNWPNumpyModalities
from ocf_datapipes.training.common import (
    DatapipeKeyForker,
    _get_datapipes_dict,
    concat_xr_time_utc,
    construct_loctime_pipelines,
    fill_nans_in_arrays,
    fill_nans_in_pv,
    normalize_gsp,
    potentially_coarsen,
    slice_datapipes_by_time,
)
from ocf_datapipes.utils.consts import (
    NWP_MEANS,
    NWP_STDS,
    RSS_MEAN,
    RSS_STD,
)
from ocf_datapipes.utils.utils import (
    combine_to_single_dataset,
    flatten_nwp_source_dict,
    nest_nwp_source_dict,
    uncombine_from_single_dataset,
)

xr.set_options(keep_attrs=True)
logger = logging.getLogger("pvnet_site_datapipe")

normalization_values = {
    2019: 3185.0,
    2020: 2678.0,
    2021: 3196.0,
    2022: 3575.0,
    2023: 3773.0,
    2024: 3773.0,
}


def normalize_pv(x: xr.DataArray):
    """Normalize PV data"""
    # This is after the data has been temporally sliced, so have the year
    return x / normalization_values[2024]

    year = x.time_utc.dt.year

    # Add the effective_capacity_mwp to the dataset, indexed on the time_utc
    return (
        x / normalization_values[year]
        if year in normalization_values
        else x / normalization_values[2024]
    )


class DictDatasetIterDataPipe(IterDataPipe):
    """Create a dictionary of xr.Datasets from a dict of datapipes"""

    datapipes_dict: dict[IterDataPipe]
    length: Optional[int]

    def __init__(self, datapipes_dict: dict[IterDataPipe]):
        """Init"""
        # Flatten the dict of input datapipes (NWP is nested)
        self.datapipes_dict = flatten_nwp_source_dict(datapipes_dict)
        self.length = None

        # Run checks
        is_okay = all([isinstance(dp, IterDataPipe) for k, dp in self.datapipes_dict.items()])

        if not is_okay:
            raise TypeError(
                "All inputs are required to be `IterDataPipe` " "for `ZipIterDataPipe`."
            )

        super().__init__()

    def __iter__(self):
        """Iter"""
        all_keys = []
        all_datapipes = []
        for k, dp in self.datapipes_dict.items():
            all_keys += [k]
            all_datapipes += [dp]

        zipped_datapipes = all_datapipes[0].zip_ocf(*all_datapipes[1:])

        for values in zipped_datapipes:
            output_dict = {key: x for key, x in zip(all_keys, values)}

            # re-nest the nwp keys
            output_dict = nest_nwp_source_dict(output_dict)

            yield output_dict


class LoadDictDatasetIterDataPipe(IterDataPipe):
    """Load NetCDF files and split them back into individual xr.Datasets"""

    filenames: List[str]
    keys: List[str]

    def __init__(self, filenames: List[str], keys: List[str]):
        """
        Load NetCDF files and split them back into individual xr.Datasets

        Args:
            filenames: List of filesnames to load
            keys: List of keys from each file to use, each key should be a
                dataarray in the xr.Dataset
        """
        super().__init__()
        self.keys = keys
        self.filenames = filenames

    def __iter__(self):
        """Iterate through each filename, loading it, uncombining it, and then yielding it"""
        while True:
            for filename in self.filenames:
                dataset = xr.open_dataset(filename)
                datasets = uncombine_from_single_dataset(dataset)
                datasets["nwp"]["ecmwf"] = potentially_coarsen(datasets["nwp"]["ecmwf"])
                # Yield a dictionary of the data, using the keys in self.keys
                dataset_dict = {}
                for k in self.keys:
                    dataset_dict[k] = datasets[k]
                yield dataset_dict


@functional_datapipe("pvnet_site_convert_to_numpy_batch")
class ConvertToNumpyBatchIterDataPipe(IterDataPipe):
    """Converts Xarray Dataset to Numpy Batch"""

    def __init__(
        self,
        dataset_dict_dp: IterDataPipe,
        check_satellite_no_zeros: bool = False,
    ):
        """Init"""
        super().__init__()
        self.dataset_dict_dp = dataset_dict_dp
        self.check_satellite_no_zeros = check_satellite_no_zeros

    def __iter__(self):
        """Iter"""
        for datapipes_dict in self.dataset_dict_dp:
            # Spatially slice, normalize, and convert data to numpy arrays
            numpy_modalities = []

            if "nwp" in datapipes_dict:
                # Combine the NWPs into NumpyBatch
                nwp_numpy_modalities = dict()
                for nwp_key, nwp_datapipe in datapipes_dict["nwp"].items():
                    nwp_numpy_modalities[nwp_key] = nwp_datapipe.convert_nwp_to_numpy_batch()

                nwp_numpy_modalities = MergeNWPNumpyModalities(nwp_numpy_modalities)
                numpy_modalities.append(nwp_numpy_modalities)

            if "sat" in datapipes_dict:
                numpy_modalities.append(datapipes_dict["sat"].convert_satellite_to_numpy_batch())
            if "pv" in datapipes_dict:
                numpy_modalities.append(datapipes_dict["pv"].convert_pv_to_numpy_batch())
            if "gsp" in datapipes_dict:
                numpy_modalities.append(datapipes_dict["gsp"].convert_gsp_to_numpy_batch())
            if "sensor" in datapipes_dict:
                numpy_modalities.append(datapipes_dict["sensor"].convert_sensor_to_numpy_batch())
            if "wind" in datapipes_dict:
                numpy_modalities.append(datapipes_dict["wind"].convert_wind_to_numpy_batch())

            logger.debug("Combine all the data sources")
            combined_datapipe = MergeNumpyModalities(numpy_modalities).add_sun_position(
                modality_name="pv"
            )

            logger.info("Filtering out samples with no data")
            # if self.check_satellite_no_zeros:
            # in production we don't want any nans in the satellite data
            #    combined_datapipe = combined_datapipe.map(check_nans_in_satellite_data)

            combined_datapipe = combined_datapipe.map(fill_nans_in_arrays)

            yield next(iter(combined_datapipe))


def minutes(num_mins: int):
    """Timedelta of a number of minutes.

    Args:
        num_mins: Minutes timedelta.
    """
    return timedelta(minutes=num_mins)


def construct_sliced_data_pipeline(
    config_filename: str,
    location_pipe: IterDataPipe,
    t0_datapipe: IterDataPipe,
    production: bool = False,
) -> dict:
    """Constructs data pipeline for the input data config file.

    This yields samples from the location and time datapipes.

    Args:
        config_filename: Path to config file.
        location_pipe: Datapipe yielding locations.
        t0_datapipe: Datapipe yielding times.
        production: Whether constucting pipeline for production inference.
    """

    datapipes_dict = _get_datapipes_dict(
        config_filename,
        production=production,
    )

    configuration = datapipes_dict.pop("config")

    # Unpack for convenience
    conf_sat = configuration.input_data.satellite
    conf_nwp = configuration.input_data.nwp

    # Slice all of the datasets by time - this is an in-place operation
    slice_datapipes_by_time(datapipes_dict, t0_datapipe, configuration, production)

    # We need a copy of the location datapipe for all keys in fork_keys
    fork_keys = set(k for k in datapipes_dict.keys())
    if "nwp" in datapipes_dict:  # NWP is nested
        fork_keys.update(set(f"nwp/{k}" for k in datapipes_dict["nwp"].keys()))

    # We don't need somes keys even if they are in the data dictionary
    fork_keys = fork_keys - set(
        ["topo", "nwp", "wind", "wind_future", "sensor", "hrv", "pv_future", "pv"]
    )

    # Set up a key-forker for all the data sources we need it for
    get_loc_datapipe = DatapipeKeyForker(fork_keys, location_pipe)

    if "nwp" in datapipes_dict:
        nwp_datapipes_dict = dict()

        for nwp_key, nwp_datapipe in datapipes_dict["nwp"].items():
            location_pipe, location_pipe_copy = location_pipe.fork(2, buffer_size=5)
            nwp_datapipe = nwp_datapipe.select_spatial_slice_pixels(
                get_loc_datapipe(f"nwp/{nwp_key}"),
                roi_height_pixels=conf_nwp[nwp_key].nwp_image_size_pixels_height,
                roi_width_pixels=conf_nwp[nwp_key].nwp_image_size_pixels_width,
            )
            # Coarsen the data, if it is separated by 0.05 degrees each
            nwp_datapipe = nwp_datapipe.map(potentially_coarsen)
            # Somewhat hacky way for India specifically, need different mean/std for ECMWF data
            if conf_nwp[nwp_key].nwp_provider in ["ecmwf"]:
                normalize_provider = "ecmwf_india"
            else:
                normalize_provider = conf_nwp[nwp_key].nwp_provider
            nwp_datapipes_dict[nwp_key] = nwp_datapipe.normalize(
                mean=NWP_MEANS[normalize_provider],
                std=NWP_STDS[normalize_provider],
            )

    if "sat" in datapipes_dict:
        sat_datapipe = datapipes_dict["sat"]

        sat_datapipe = sat_datapipe.select_spatial_slice_pixels(
            get_loc_datapipe("sat"),
            roi_height_pixels=conf_sat.satellite_image_size_pixels_height,
            roi_width_pixels=conf_sat.satellite_image_size_pixels_width,
        )
        sat_datapipe = sat_datapipe.normalize(mean=RSS_MEAN, std=RSS_STD)

    if "pv" in datapipes_dict:
        # Recombine Sensor arrays - see function doc for further explanation
        pv_datapipe = (
            datapipes_dict["pv"].zip_ocf(datapipes_dict["pv_future"]).map(concat_xr_time_utc)
        )
        pv_datapipe = pv_datapipe.normalize(normalize_fn=normalize_pv)
        pv_datapipe = pv_datapipe.map(fill_nans_in_pv)

    finished_dataset_dict = {"config": configuration}

    if "gsp" in datapipes_dict:
        gsp_future_datapipe = datapipes_dict["gsp_future"]
        gsp_future_datapipe = gsp_future_datapipe.select_spatial_slice_meters(
            location_datapipe=get_loc_datapipe("gsp_future"),
            roi_height_meters=1,
            roi_width_meters=1,
            dim_name="gsp_id",
        )

        gsp_datapipe = datapipes_dict["gsp"]
        gsp_datapipe = gsp_datapipe.select_spatial_slice_meters(
            location_datapipe=get_loc_datapipe("gsp"),
            roi_height_meters=1,
            roi_width_meters=1,
            dim_name="gsp_id",
        )

        # Recombine GSP arrays - see function doc for further explanation
        gsp_datapipe = gsp_datapipe.zip_ocf(gsp_future_datapipe).map(concat_xr_time_utc)
        gsp_datapipe = gsp_datapipe.normalize(normalize_fn=normalize_gsp)

        finished_dataset_dict["gsp"] = gsp_datapipe

    get_loc_datapipe.close()

    if "nwp" in datapipes_dict:
        finished_dataset_dict["nwp"] = nwp_datapipes_dict
    if "sat" in datapipes_dict:
        finished_dataset_dict["sat"] = sat_datapipe
    if "pv" in datapipes_dict:
        finished_dataset_dict["pv"] = pv_datapipe

    return finished_dataset_dict


def pvnet_site_datapipe(
    config_filename: str,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
) -> IterDataPipe:
    """
    Construct PVNet site pipeline for the input data config file.

    Args:
        config_filename: Path to config file.
        start_time: Minimum time at which a sample can be selected.
        end_time: Maximum time at which a sample can be selected.
    """
    logger.info("Constructing pvnet site pipeline")

    # Open datasets from the config and filter to useable location-time pairs
    location_pipe, t0_datapipe = construct_loctime_pipelines(
        config_filename,
        start_time,
        end_time,
    )

    # Shard after we have the loc-times. These are already shuffled so no need to shuffle again
    location_pipe = location_pipe.sharding_filter()
    t0_datapipe = t0_datapipe.sharding_filter()

    # In this function we re-open the datasets to make a clean separation before/after sharding
    # This function
    datapipe_dict = construct_sliced_data_pipeline(
        config_filename,
        location_pipe,
        t0_datapipe,
    )

    # Merge all the datapipes into one
    return DictDatasetIterDataPipe(
        {k: v for k, v in datapipe_dict.items() if k != "config"},
    ).map(combine_to_single_dataset)


def split_dataset_dict_dp(element):
    """
    Wrap each of the data source inputs into a datapipe

    Args:
        element: Dictionary of xarray objects
    """

    element = flatten_nwp_source_dict(element)
    output_dict = {k: IterableWrapper([v]) for k, v in element.items() if k != "config"}
    output_dict = nest_nwp_source_dict(output_dict)

    return output_dict


def pvnet_site_netcdf_datapipe(
    keys: List[str],
    filenames: List[str],
) -> IterDataPipe:
    """
    Load the saved Datapipes from pvnet site, and transform to numpy batch

    Args:
        keys: List of keys to extract from the single NetCDF files
        filenames: List of NetCDF files to load

    Returns:
        Datapipe that transforms the NetCDF files to numpy batch
    """
    logger.info("Constructing pvnet site file pipeline")
    # Load files
    datapipe_dict_dp: IterDataPipe = LoadDictDatasetIterDataPipe(
        filenames=filenames,
        keys=keys,
    ).map(split_dataset_dict_dp)
    datapipe = datapipe_dict_dp.pvnet_site_convert_to_numpy_batch()

    return datapipe


if __name__ == "__main__":
    # Load the saved NetCDF files here
    datapipe = pvnet_site_netcdf_datapipe(
        keys=["nwp", "pv"],
        filenames=["/run/media/jacob/data/train/000000.nc"],
    )
    batch = next(iter(datapipe))
    # print(batch)
    # print(len(batch[BatchKey.pv_solar_azimuth]))
    print(batch[BatchKey.pv_solar_elevation].shape)



"""Load Metoematics data from local files for training/inference"""

import logging

import numpy as np
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.config.model import Sensor

_log = logging.getLogger(__name__)


@functional_datapipe("OpenMeteomatics")
class OpenMeteomaticsFromZarrIterDataPipe(IterDataPipe):
    """OpenMeteomaticsFromNetCDFIterDataPipe"""

    def __init__(
        self,
        sensor: Sensor,
    ):
        """
        Datapipe to load Meteomatics point data

        Args:
            sensor: Sensor configuration
        """
        super().__init__()
        self.sensor = sensor
        self.filename = self.sensor.sensor_filename
        self.variables = list(self.sensor.sensor_variables)

    def __iter__(self):
        if len(self.filename) == 1 or "*" not in self.filename:
            data = xr.open_zarr(self.filename)
        else:
            data = xr.open_mfdataset(
                self.filename, engine="zarr", combine="nested", concat_dim="time_utc"
            )
        data = data.sortby("time_utc")
        if "wind_speed_100m:ms" in data:
            # Generate U and V components from wind speed and direction
            data["100u"] = data["wind_speed_100m:ms"] * np.cos(np.deg2rad(data["wind_dir_100m:d"]))
            data["100v"] = data["wind_speed_100m:ms"] * np.sin(np.deg2rad(data["wind_dir_100m:d"]))
            data["10u"] = data["wind_speed_10m:ms"] * np.cos(np.deg2rad(data["wind_dir_10m:d"]))
            data["10v"] = data["wind_speed_10m:ms"] * np.sin(np.deg2rad(data["wind_dir_10m:d"]))
            data["200u"] = data["wind_speed_200m:ms"] * np.cos(np.deg2rad(data["wind_dir_200m:d"]))
            data["200v"] = data["wind_speed_200m:ms"] * np.sin(np.deg2rad(data["wind_dir_200m:d"]))
        data = data[self.variables].to_array()
        while True:
            yield data



"""Convert PV to Numpy Batch"""

import logging

import numpy as np
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.batch import BatchKey, NumpyBatch
from ocf_datapipes.utils.utils import datetime64_to_float

logger = logging.getLogger(__name__)


def convert_sensor_to_numpy_batch(xr_data):
    """Convert Sensor Xarray to NumpyBatch"""

    example: NumpyBatch = {
        BatchKey.sensor: xr_data.values,
        BatchKey.sensor_t0_idx: xr_data.attrs["t0_idx"],
        BatchKey.sensor_id: xr_data["station_id"].values.astype(np.float32),
        # BatchKey.sensor_observed_capacity_wp: (xr_data["observed_capacity_wp"].values),
        # BatchKey.sensor_nominal_capacity_wp: (xr_data["nominal_capacity_wp"].values),
        BatchKey.sensor_time_utc: datetime64_to_float(xr_data["time_utc"].values),
        BatchKey.sensor_latitude: xr_data["latitude"].values,
        BatchKey.sensor_longitude: xr_data["longitude"].values,
    }
    return example


@functional_datapipe("convert_sensor_to_numpy_batch")
class ConvertSensorToNumpyBatchIterDataPipe(IterDataPipe):
    """Convert Sensor Xarray to NumpyBatch"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Convert sensor Xarray objects to NumpyBatch objects

        Args:
            source_datapipe: Datapipe emitting sensor Xarray objects
        """
        super().__init__()
        self.source_datapipe = source_datapipe

    def __iter__(self) -> NumpyBatch:
        """Iterate and convert sensor Xarray to NumpyBatch"""
        for xr_data in self.source_datapipe:
            yield convert_sensor_to_numpy_batch(xr_data)



"""Remove any data with nans"""

import logging

import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("select_non_nan_times")
class SelectNonNaNTimesIterDataPipe(IterDataPipe):
    """Remove timestamps where any value in the xarray object is NaN"""

    def __init__(self, source_datapipe: IterDataPipe, time_dim: str = "time_utc"):
        """
        Remove timestamps where any value in the dataset is NaN

        Args:
            source_datapipe: Datapipe of PV data
            time_dim: the time dimension to drop nans along
        """
        self.source_datapipe = source_datapipe
        self.time_dim = time_dim

    def __iter__(self) -> xr.DataArray:
        for xr_data in self.source_datapipe:
            logger.debug(
                f"Dropping nans on {self.time_dim}. "
                f"Currently there are {len(xr_data[self.time_dim])}"
            )
            xr_data_new = xr_data.dropna(dim=self.time_dim)
            if len(xr_data_new[self.time_dim]) == 0:
                logger.debug(xr_data)
                raise Exception("Data has only nans in it")
            logger.debug(
                f"After dropping nans on {self.time_dim}, "
                f"there are {len(xr_data_new[self.time_dim])}"
            )
            yield xr_data_new



""" Util functions for PV data source"""

from typing import Optional

import numpy as np
import pandas as pd
import xarray as xr


def put_pv_data_into_an_xr_dataarray(
    df_gen: pd.DataFrame,
    observed_system_capacities: pd.Series,
    nominal_system_capacities: pd.Series,
    ml_id: pd.Series,
    longitude: pd.Series,
    latitude: pd.Series,
    tilt: Optional[pd.Series] = None,
    orientation: Optional[pd.Series] = None,
) -> xr.DataArray:
    """Convert to an xarray DataArray.

    Args:
        df_gen: pd.DataFrame where the columns are PV systems (and the column names are ints), and
            the index is UTC datetime
        observed_system_capacities: The max power output observed in the time series for PV system
            in watts. Index is PV system IDs
        nominal_system_capacities: The metadata value for each PV system capacities in watts
        ml_id: The `ml_id` used to identify each PV system
        longitude: longitude of the locations
        latitude: latitude of the locations
        tilt: Tilt of the panels
        orientation: Orientation of the panels
    """
    # Sanity check!
    system_ids = df_gen.columns
    for name, series in (
        ("observed_system_capacities", observed_system_capacities),
        ("nominal_system_capacities", nominal_system_capacities),
        ("ml_id", ml_id),
        ("longitude", longitude),
        ("latitude", latitude),
        ("tilt", tilt),
        ("orientation", orientation),
    ):
        if (series is not None) and (not np.array_equal(series.index, system_ids)):
            raise ValueError(
                f"Index of {name} does not equal {system_ids}. Index is {series.index}"
            )

    data_array = xr.DataArray(
        data=df_gen.values,
        coords=(
            ("time_utc", df_gen.index.values),
            ("pv_system_id", system_ids),
        ),
        name="pv_power_watts",
    ).astype(np.float32)

    data_array = data_array.assign_coords(
        observed_capacity_wp=("pv_system_id", observed_system_capacities),
        nominal_capacity_wp=("pv_system_id", nominal_system_capacities),
        ml_id=("pv_system_id", ml_id),
        longitude=("pv_system_id", longitude),
        latitude=("pv_system_id", latitude),
    )

    if tilt is not None:
        data_array = data_array.assign_coords(
            tilt=("pv_system_id", tilt),
        )
    if orientation is not None:
        data_array = data_array.assign_coords(
            orientation=("pv_system_id", orientation),
        )

    return data_array



"""Selects time slice"""

import logging
from datetime import timedelta
from typing import Optional, Union

import numpy as np
import pandas as pd
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


def fill_1d_bool_gaps(x, max_gap, fill_ends=False):
    """In a boolean array, fill consecutive False elements if their number is less than the gap_size

    Args:
        x: A 1-dimensional boolean array
        max_gap: integer of the maximum gap size which will be filled with True
        fill_ends: Whether to fill the ends as if there are True values on either side

    Returns:
        A 1-dimensional boolean array

    Examples:
        >>> x = np.array([0, 1, 0, 0, 1, 0, 1, 0])
        >>> fill_1d_bool_gaps(x, max_gap=2, fill_ends=False).astype(int)
        array([0, 1, 1, 1, 1, 1, 1, 0])

        >>> x = np.array([0, 1, 0, 0, 1, 0, 1, 0])
        >>> fill_1d_bool_gaps(x, max_gap=1, fill_ends=True).astype(int)
        array([1, 1, 0, 0, 1, 1, 1, 1])
    """
    if fill_ends:
        x_extended = np.concatenate([[True], x, [True]])
        return fill_1d_bool_gaps(x_extended, max_gap, fill_ends=False)[1:-1]

    should_fill = np.zeros(len(x), dtype=bool)

    i_start = None

    last_b = False
    for i, b in enumerate(x):
        if last_b and not b:
            i_start = i
        elif b and not last_b and i_start is not None:
            if i - i_start <= max_gap:
                should_fill[i_start:i] = True
            i_start = None
        last_b = b

    return np.logical_or(should_fill, x)


@functional_datapipe("select_time_slice")
class SelectTimeSliceIterDataPipe(IterDataPipe):
    """Selects time slice"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        t0_datapipe: IterDataPipe,
        sample_period_duration: timedelta,
        history_duration: Optional[timedelta] = None,
        forecast_duration: Optional[timedelta] = None,
        interval_start: Optional[timedelta] = None,
        interval_end: Optional[timedelta] = None,
        fill_selection: Optional[bool] = False,
        max_steps_gap: Optional[int] = 0,
    ):
        """
        Selects time slice.

        Either `history_duration` and `history_duration` or `interval_start` and
        `interval_end` must be supplied.

        Args:
            source_datapipe: Datapipe of Xarray objects
            t0_datapipe: Datapipe of t0 times
            sample_period_duration: Sample period of xarray data
            history_duration (optional): History time used
            forecast_duration (optional): Forecast time used
            interval_start (optional): timedelta with respect to t0 where the open interval begins
            interval_end (optional): timedelta with respect to t0 where the open interval ends
            fill_selection (optional): If True, and if the data yielded from `source_datapipe` does
                not extend over the entire requested time period. The missing timestamps are filled
                in the returned xarray object tyo give the expected shape. Else the default xarray
                slicing behaviour is used and timestamps may be missing. When filled, the values are
                linearly interpolated up to a gap size of `max_steps_gap` steps. If outside this
                range, the values are set to NaN.
            max_steps_gap (optional): The number of consecutive missing time steps which will be
                filled via linear interpolation. If set to zero, no interpolation is used and all
                missing timesteps will be NaN.
        """
        self.source_datapipe = source_datapipe
        self.t0_datapipe = t0_datapipe
        self.fill_selection = fill_selection
        self.max_steps_gap = max_steps_gap

        used_duration = history_duration is not None and forecast_duration is not None
        used_intervals = interval_start is not None and interval_end is not None
        assert used_duration ^ used_intervals, "Either durations, or intervals must be supplied"
        assert max_steps_gap >= 0, "max_steps_gap must be >= 0 "

        if used_duration:
            self.interval_start = -np.timedelta64(history_duration)
            self.interval_end = np.timedelta64(forecast_duration)
        elif used_intervals:
            self.interval_start = np.timedelta64(interval_start)
            self.interval_end = np.timedelta64(interval_end)

        self.sample_period_duration = sample_period_duration

        if self.fill_selection and max_steps_gap == 0:
            self._sel = self._sel_fillnan
        elif self.fill_selection and max_steps_gap > 0:
            self._sel = self._sel_fillinterp
        else:
            self._sel = self._sel_default

    def _sel_fillnan(self, xr_data, start_dt, end_dt):
        requested_times = pd.date_range(
            start_dt,
            end_dt,
            freq=self.sample_period_duration,
        )
        # Missing time indexes are returned with all NaN values
        return xr_data.reindex(time_utc=requested_times)

    def _sel_fillinterp(self, xr_data, start_dt, end_dt):
        dt_buffer = self.sample_period_duration * self.max_steps_gap

        # Initially select larger period so we can use it to interpolate to requested period
        # This slice also avoids us interpolating the whole dataset to get the requested times
        ds = xr_data.sel(time_utc=slice(start_dt - dt_buffer, end_dt + dt_buffer))

        # These are the times we will ultimately return
        requested_times = pd.date_range(
            start_dt,
            end_dt,
            freq=self.sample_period_duration,
        )

        # These are the times we use for interpolation to the requested_times
        buffer_requested_times = pd.date_range(
            start_dt - dt_buffer,
            end_dt + dt_buffer,
            freq=self.sample_period_duration,
        )

        # If all the requested times are present we avoid running interpolation
        if np.isin(requested_times, ds.time_utc).all():
            return ds.sel(time_utc=slice(start_dt, end_dt))

        # If less than 2 of the buffer requested times are present we cannot infill
        elif np.isin(buffer_requested_times, ds.time_utc).sum() < 2:
            logger.warning("Cannot run interpolate infilling with less than 2 time steps available")
            return self._sel_fillnan(xr_data, start_dt, end_dt)

        logger.info("Some requested times are missing - running interpolation")
        # Find the timestamps which are within max gap size
        mask = np.isin(buffer_requested_times, ds.time_utc)
        valid_fill_times = fill_1d_bool_gaps(mask, self.max_steps_gap, fill_ends=False)

        # Run the interpolation and filter to requested times
        ds_interp = ds.interp(time_utc=buffer_requested_times, method="linear", assume_sorted=True)

        # Mask the timestamps outside the max gap size
        valid_fill_times_xr = xr.zeros_like(ds_interp.time_utc, dtype=bool)
        valid_fill_times_xr.values[:] = valid_fill_times

        valid_requested_times = valid_fill_times_xr.sel(time_utc=slice(start_dt, end_dt))
        if not valid_requested_times.all():
            not_infilled_times = valid_requested_times.where(~valid_requested_times, drop=True)
            logger.warning(
                "After interpolation the following requested times are still missing:"
                f"{not_infilled_times.time_utc.values}"
            )

        ds_out = ds_interp.where(valid_fill_times_xr)

        # Filter to selected times
        ds_out = ds_out.sel(time_utc=slice(start_dt, end_dt))
        return ds_out

    def _sel_default(self, xr_data, start_dt, end_dt):
        return xr_data.sel(time_utc=slice(start_dt, end_dt))

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        for t0, xr_data in self.t0_datapipe.zip(self.source_datapipe):
            t0_datetime_utc = pd.Timestamp(t0)
            start_dt = t0_datetime_utc + self.interval_start
            end_dt = t0_datetime_utc + self.interval_end

            start_dt = start_dt.ceil(self.sample_period_duration)
            end_dt = end_dt.ceil(self.sample_period_duration)

            yield self._sel(xr_data, start_dt, end_dt)



"""Convert Satellite to NumpyBatch"""

from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.batch import BatchKey, NumpyBatch
from ocf_datapipes.utils.utils import datetime64_to_float


def _convert_satellite_to_numpy_batch(xr_data):
    example: NumpyBatch = {
        BatchKey.satellite_actual: xr_data.values,
        BatchKey.satellite_t0_idx: xr_data.attrs["t0_idx"],
        BatchKey.satellite_time_utc: datetime64_to_float(xr_data["time_utc"].values),
    }

    for batch_key, dataset_key in (
        (BatchKey.satellite_y_geostationary, "y_geostationary"),
        (BatchKey.satellite_x_geostationary, "x_geostationary"),
    ):
        # HRVSatellite coords are already float32.
        example[batch_key] = xr_data[dataset_key].values

    return example


def _convert_hrvsatellite_to_numpy_batch(xr_data):
    example: NumpyBatch = {
        BatchKey.hrvsatellite_actual: xr_data.values,
        BatchKey.hrvsatellite_t0_idx: xr_data.attrs["t0_idx"],
        BatchKey.hrvsatellite_time_utc: datetime64_to_float(xr_data["time_utc"].values),
    }

    for batch_key, dataset_key in (
        (BatchKey.hrvsatellite_y_geostationary, "y_geostationary"),
        (BatchKey.hrvsatellite_x_geostationary, "x_geostationary"),
    ):
        # Satellite coords are already float32.
        example[batch_key] = xr_data[dataset_key].values

    return example


def convert_satellite_to_numpy_batch(xr_data, is_hrv=False):
    """Converts Xarray Satellite to NumpyBatch object"""
    if is_hrv:
        example = _convert_hrvsatellite_to_numpy_batch(xr_data)
    else:
        example = _convert_satellite_to_numpy_batch(xr_data)
    return example


@functional_datapipe("convert_satellite_to_numpy_batch")
class ConvertSatelliteToNumpyBatchIterDataPipe(IterDataPipe):
    """Converts Xarray Satellite to NumpyBatch object"""

    def __init__(self, source_datapipe: IterDataPipe, is_hrv: bool = False):
        """
        Converts Xarray satellite object to NumpyBatch object

        Args:
            source_datapipe: Datapipe emitting Xarray satellite objects
            is_hrv: Whether this is HRV satellite data or non-HRV data
        """
        super().__init__()
        self.source_datapipe = source_datapipe
        self.is_hrv = is_hrv

    def __iter__(self) -> NumpyBatch:
        """Convert each example to a NumpyBatch object"""
        for xr_data in self.source_datapipe:
            yield convert_satellite_to_numpy_batch(xr_data, self.is_hrv)



"""GSP Loader"""

import datetime
import logging
from pathlib import Path
from typing import Optional, Union

import numpy as np
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.load.gsp.utils import get_gsp_id_to_shape, put_gsp_data_into_an_xr_dataarray

logger = logging.getLogger(__name__)


@functional_datapipe("open_gsp")
class OpenGSPIterDataPipe(IterDataPipe):
    """Get and open the GSP data"""

    def __init__(
        self,
        gsp_pv_power_zarr_path: Union[str, Path],
        gsp_id_to_region_id_filename: Optional[str] = None,
        sheffield_solar_region_path: Optional[str] = None,
        threshold_mw: int = 0,
        sample_period_duration: datetime.timedelta = datetime.timedelta(minutes=30),
    ):
        """
        Get and open the GSP data

        Args:
            gsp_pv_power_zarr_path: Path to the Zarr for GSP PV Power
            gsp_id_to_region_id_filename: Path to the file containing the mapping of ID ot region
            sheffield_solar_region_path: Path to the Sheffield Solar region data
            threshold_mw: Threshold to drop GSPs by
            sample_period_duration: Sample period of the GSP data
        """
        self.gsp_pv_power_zarr_path = gsp_pv_power_zarr_path

        self.gsp_id_to_region_id_filename = gsp_id_to_region_id_filename
        self.sheffield_solar_region_path = sheffield_solar_region_path
        self.threshold_mw = threshold_mw
        self.sample_period_duration = sample_period_duration

    def __iter__(self) -> xr.DataArray:
        """Get and return GSP data"""
        gsp_id_to_shape = get_gsp_id_to_shape(
            self.gsp_id_to_region_id_filename,
            self.sheffield_solar_region_path,
        )

        logger.debug(f"Getting GSP data from {self.gsp_pv_power_zarr_path}")

        # Load GSP generation xr.Dataset
        gsp_pv_power_mw_ds = xr.open_dataset(self.gsp_pv_power_zarr_path, engine="zarr")

        # Ensure the centroids have the same GSP ID index as the GSP PV power
        gsp_id_to_shape = gsp_id_to_shape.loc[gsp_pv_power_mw_ds.gsp_id]

        data_array = put_gsp_data_into_an_xr_dataarray(
            gsp_pv_power_mw=gsp_pv_power_mw_ds.generation_mw.data.astype(np.float32),
            time_utc=gsp_pv_power_mw_ds.datetime_gmt.data,
            gsp_id=gsp_pv_power_mw_ds.gsp_id.data,
            # TODO: Try using `gsp_id_to_shape.geometry.envelope.centroid`. See issue #76.
            x_osgb=gsp_id_to_shape.x_osgb.astype(np.float32),
            y_osgb=gsp_id_to_shape.y_osgb.astype(np.float32),
            nominal_capacity_mwp=gsp_pv_power_mw_ds.installedcapacity_mwp.data.astype(np.float32),
            effective_capacity_mwp=gsp_pv_power_mw_ds.capacity_mwp.data.astype(np.float32),
        )

        del gsp_id_to_shape, gsp_pv_power_mw_ds
        while True:
            yield data_array



"""Additional utils for working with batches"""

import numpy as np
import torch

from ocf_datapipes.batch import NumpyBatch, TensorBatch


def _copy_batch_to_device(batch: dict, device: torch.device) -> dict:
    """
    Moves tensor leaves in a nested dict to a new device

    Args:
        batch: nested dict with tensors to move
        device: Device to move tensors to

    Returns:
        A dict with tensors moved to new device
    """
    batch_copy = {}

    for k, v in batch.items():
        if isinstance(v, dict):
            # Recursion to reach the nested NWP
            batch_copy[k] = _copy_batch_to_device(v, device)
        elif isinstance(v, torch.Tensor):
            batch_copy[k] = v.to(device)
        else:
            batch_copy[k] = v
    return batch_copy


def copy_batch_to_device(batch: TensorBatch, device: torch.device) -> TensorBatch:
    """
    Moves the tensors in a TensorBatch to a new device.

    Args:
        batch: TensorBatch with tensors to move
        device: Device to move tensors to

    Returns:
        TensorBatch with tensors moved to new device
    """
    return _copy_batch_to_device(batch, device)


def _batch_to_tensor(batch: dict) -> dict:
    """
    Moves ndarrays in a nested dict to torch tensors

    Args:
        batch: nested dict with data in numpy arrays

    Returns:
        Nested dict with data in torch tensors
    """
    for k, v in batch.items():
        if isinstance(v, dict):
            # Recursion to reach the nested NWP
            batch[k] = _batch_to_tensor(v)
        elif isinstance(v, np.ndarray) and np.issubdtype(v.dtype, np.number):
            batch[k] = torch.as_tensor(v)
    return batch


def batch_to_tensor(batch: NumpyBatch) -> TensorBatch:
    """
    Moves data in a NumpyBatch to a TensorBatch

    Args:
        batch: NumpyBatch with data in numpy arrays

    Returns:
        TensorBatch with data in torch tensors
    """
    return _batch_to_tensor(batch)



"""Datapipes to add Sun position to NumpyBatch"""

import warnings

import numpy as np
import pvlib
from torch.utils.data import IterDataPipe, functional_datapipe

from ocf_datapipes.batch import BatchKey
from ocf_datapipes.utils.consts import (
    AZIMUTH_MEAN,
    AZIMUTH_STD,
    ELEVATION_MEAN,
    ELEVATION_STD,
)
from ocf_datapipes.utils.geospatial import osgb_to_lon_lat


def _get_azimuth_and_elevation(lon, lat, dt, must_be_finite):
    if type(dt[0]) == np.datetime64:
        # This caused an issue if it was 'datetime64[s]'
        dt = np.array(dt, dtype="datetime64[ns]")

    if not np.isfinite([lon, lat]).all():
        if must_be_finite:
            raise ValueError(f"Non-finite (lon, lat) = ({lon}, {lat}")
        return (
            np.full_like(dt, fill_value=np.nan).astype(np.float32),
            np.full_like(dt, fill_value=np.nan).astype(np.float32),
        )

    else:
        solpos = pvlib.solarposition.get_solarposition(
            time=dt,
            latitude=lat,
            longitude=lon,
            # Which `method` to use?
            # pyephem seemed to be a good mix between speed and ease
            # but causes segfaults!
            # nrel_numba doesn't work when using multiple worker processes.
            # nrel_c is probably fastest but requires C code to be
            #   manually compiled: https://midcdmz.nrel.gov/spa/
        )
        azimuth = solpos["azimuth"]
        elevation = solpos["elevation"]
        return azimuth, elevation


@functional_datapipe("add_sun_position")
class AddSunPositionIterDataPipe(IterDataPipe):
    """Adds the sun position to the NumpyBatch"""

    def __init__(self, source_datapipe: IterDataPipe, modality_name: str):
        """
        Adds the sun position to the NumpyBatch

        Args:
            source_datapipe: Datapipe of NumpyBatch
            modality_name: Modality to add the sun position for
        """
        self.source_datapipe = source_datapipe
        self.modality_name = modality_name
        assert self.modality_name in [
            "hrvsatellite",
            "gsp",
            "pv",
            "wind",
        ], f"Cant add sun position on {self.modality_name}"

    def __iter__(self):
        for np_batch in self.source_datapipe:
            if self.modality_name == "hrvsatellite":
                # TODO Make work with Lat/Lons instead
                y_osgb = np_batch[BatchKey.hrvsatellite_y_osgb]  # Shape: optional[example], y, x
                x_osgb = np_batch[BatchKey.hrvsatellite_x_osgb]  # Shape: optional[example], y, x
                time_utc = np_batch[
                    BatchKey.hrvsatellite_time_utc
                ]  # Shape: optional[example], time

                # Get the time and position for the centre of the t0 frame:
                y_centre_idx = int(y_osgb.shape[-2] // 2)
                x_centre_idx = int(y_osgb.shape[-1] // 2)
                y_osgb = y_osgb[..., y_centre_idx, x_centre_idx]  # Shape: optional[example]
                x_osgb = x_osgb[..., y_centre_idx, x_centre_idx]  # Shape: optional[example]

            elif self.modality_name == "pv":
                lats = np_batch[BatchKey.pv_latitude]  # Shape: (optional[example], n_pvs)
                lons = np_batch[BatchKey.pv_longitude]  # Shape: (optional[example], n_pvs)
                time_utc = np_batch[BatchKey.pv_time_utc]  # Shape: optional[example]

                # If using multiple PV systems, take mean location
                if lats.shape[-1] > 1:
                    # Sometimes, the PV coords can all be NaNs if there are no PV systems
                    # for that datetime and location. e.g. in northern Scotland!
                    with warnings.catch_warnings():
                        warnings.filterwarnings(
                            action="ignore", category=RuntimeWarning, message="Mean of empty slice"
                        )
                        lats = np.nanmean(lats, axis=-1)
                        lons = np.nanmean(lons, axis=-1)
                else:
                    lats = lats[..., 0]
                    lons = lons[..., 0]

            elif self.modality_name == "wind":
                lats = np_batch[BatchKey.wind_latitude]  # Shape: (optional[example], n_pvs)
                lons = np_batch[BatchKey.wind_longitude]  # Shape: (optional[example], n_pvs)
                time_utc = np_batch[BatchKey.wind_time_utc]  # Shape: optional[example]

                # If using multiple PV systems, take mean location
                if lats.shape[-1] > 1:
                    # Sometimes, the PV coords can all be NaNs if there are no PV systems
                    # for that datetime and location. e.g. in northern Scotland!
                    with warnings.catch_warnings():
                        warnings.filterwarnings(
                            action="ignore", category=RuntimeWarning, message="Mean of empty slice"
                        )
                        lats = np.nanmean(lats, axis=-1)
                        lons = np.nanmean(lons, axis=-1)
                else:
                    lats = lats[..., 0]
                    lons = lons[..., 0]

            elif self.modality_name == "gsp":
                y_osgb = np_batch[BatchKey.gsp_y_osgb]  # Shape: optional[example], n_gsps
                x_osgb = np_batch[BatchKey.gsp_x_osgb]  # Shape: optional[example], n_gsps
                time_utc = np_batch[BatchKey.gsp_time_utc]  # Shape: optional[example],

                # If using multiple GSPs, take mean location
                if x_osgb.shape[-1] > 1:
                    y_osgb = np.nanmean(y_osgb, axis=-1)
                    x_osgb = np.nanmean(x_osgb, axis=-1)
                else:
                    y_osgb = y_osgb[..., 0]
                    x_osgb = x_osgb[..., 0]

            else:
                raise ValueError(f"Unrecognized modality: {self.modality_name }")

            # As we move away from OSGB and towards lon, lat we can exclude more sources here
            if self.modality_name not in ["pv", "wind"]:
                # Convert to the units that pvlib expects: lon, lat
                lons, lats = osgb_to_lon_lat(x=x_osgb, y=y_osgb)

            # Elevations must be finite and non-nan except for PV data where values may be missing
            must_be_finite = self.modality_name != "pv"

            # Check if the input is batched
            # time_utc could have shape (batch_size, n_times) or (n_times,)
            assert len(time_utc.shape) in [1, 2]
            is_batched = len(time_utc.shape) == 2

            times = time_utc.astype("datetime64[s]")

            if is_batched:
                assert lons.shape == (time_utc.shape[0],)
                assert lats.shape == (time_utc.shape[0],)

                azimuth = np.full_like(time_utc, fill_value=np.nan).astype(np.float32)
                elevation = np.full_like(time_utc, fill_value=np.nan).astype(np.float32)

                # Loop round each example to get the Sun's elevation and azimuth
                for example_idx, (lon, lat, dt) in enumerate(zip(lons, lats, times)):
                    azimuth[example_idx], elevation[example_idx] = _get_azimuth_and_elevation(
                        lon, lat, dt, must_be_finite
                    )
            else:
                assert (isinstance(lons, np.ndarray) and lons.shape == ()) or isinstance(
                    lons, float
                )
                assert (isinstance(lats, np.ndarray) and lats.shape == ()) or isinstance(
                    lats, float
                )

                azimuth, elevation = _get_azimuth_and_elevation(lons, lats, times, must_be_finite)

            # Normalize
            azimuth = (azimuth - AZIMUTH_MEAN) / AZIMUTH_STD
            elevation = (elevation - ELEVATION_MEAN) / ELEVATION_STD

            # Store
            azimuth_batch_key = BatchKey[self.modality_name + "_solar_azimuth"]
            elevation_batch_key = BatchKey[self.modality_name + "_solar_elevation"]
            np_batch[azimuth_batch_key] = azimuth
            np_batch[elevation_batch_key] = elevation

            yield np_batch



"""Xarray transforms"""

from .add_t0idx_and_sample_period_duration import (
    AddT0IdxAndSamplePeriodDurationIterDataPipe as AddT0IdxAndSamplePeriodDuration,
)
from .convert_pressure_levels_to_separate_variables import (
    ConvertPressureLevelsToSeparateVariablesIterDataPipe as ConvertPressureLevelsToSeparateVariables,  # noqa: E501
)
from .create_sun_image import CreateSunImageIterDataPipe as CreateSunImage
from .create_time_image import CreateTimeImageIterDataPipe as CreateTimeImage
from .downsample import DownsampleIterDataPipe as Downsample
from .gsp.create_gsp_image import CreateGSPImageIterDataPipe as CreateGSPImage
from .gsp.ensure_n_gsp_per_example import (
    EnsureNGSPSPerExampleIterDataPipe as EnsureNGSPSPerExampleIter,
)
from .normalize import NormalizeIterDataPipe as Normalize
from .pv.assign_daynight_status import (
    AssignDayNightStatusIterDataPipe as AssignDayNightStatus,
)
from .pv.create_pv_image import CreatePVImageIterDataPipe as CreatePVImage
from .pv.create_pv_meta_image import CreatePVMetadataImageIterDataPipe as CreatePVMetadataImage
from .pv.ensure_n_pv_systems_per_example import (
    EnsureNPVSystemsPerExampleIterDataPipe as EnsureNPVSystemsPerExample,
)
from .pv.pv_fill_nighttime_nans_with_zeros import PVFillNightNansIterDataPipe as PVFillNightNans
from .pv.pv_infill_interpolate import PVInterpolateInfillIterDataPipe as PVInterpolateInfill
from .pv.pv_power_rolling_window import PVPowerRollingWindowIterDataPipe as PVPowerRollingWindow
from .pv.remove_pv_zero_examples import PVPowerRemoveZeroDataIterDataPipe as PVPowerRemoveZeroData
from .reproject_topographic_data import (
    ReprojectTopographyIterDataPipe as ReprojectTopography,
)
from .upsample import UpSampleIterDataPipe as Upsample



"""Convert NWP data to the target time with dropout"""

import logging
from datetime import timedelta
from typing import List, Optional, Union

import numpy as np
import pandas as pd
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("apply_pv_dropout")
class ApplyPVDropoutIterDataPipe(IterDataPipe):
    """Apply PV system dropout to mimic production

    Systems have independent delay times. Systems may also completely drop out.

    """

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        min_frac: float,
        max_frac: float,
        system_dropout_timedeltas: Optional[List[timedelta]],
    ):
        """Apply PV system dropout to mimic production

        Systems have independent delay times. Systems may also completely drop out.

        For each yielded sample, a dropout fraction will be randomly chosen from the range
        [min_frac, max_frac]. This random fraction will be used as the chance of dropout for each
        PV system in the sample. Using a list instead of a single value allows us to avoid
        overfitting to the fraction of dropped out systems.

        Args:
            source_datapipe: Datapipe emitting an Xarray Dataset with time_utc indexer.
            min_frac: The minimum chance for each system to be dropped out in a sample
            max_frac: The maximum chance for each system to be dropped out in a sample
            system_dropout_timedeltas: List of timedeltas. We randomly select the delay for each PV
                system from this list. These should be negative timedeltas w.r.t the last time_utc
                value of the xarray data.
        """
        self.source_datapipe = source_datapipe
        self.min_frac = min_frac
        self.max_frac = max_frac
        self.system_dropout_timedeltas = system_dropout_timedeltas

        if system_dropout_timedeltas is not None:
            assert (
                len(system_dropout_timedeltas) >= 1
            ), "Must include list of relative dropout timedeltas"

            assert all(
                [t <= timedelta(minutes=0) for t in system_dropout_timedeltas]
            ), f"dropout timedeltas must be negative: {system_dropout_timedeltas}"

        assert all(
            [0 <= f <= 1 for f in [min_frac, max_frac]]
        ), "The min and max dropout fractions must be in open range (0, 1)"

        assert min_frac <= max_frac, "Min dropout fraction <= maximum dropout fraction"

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """Iterate through Xarray dataset using dropout"""

        for xr_data in self.source_datapipe:
            # Assign these values for convenience
            t0 = pd.Timestamp(xr_data.time_utc.values[-1])
            n_systems = len(xr_data.pv_system_id)

            if not (self.min_frac == self.max_frac == 0):
                # Apply PV system dropout - individual systems are dropped out

                # Don't want fraction of dropped out system to be the same in each sample
                # This might lead to overfitting. Instead vary it
                dropout_p = np.random.uniform(low=self.min_frac, high=self.max_frac)

                system_mask = xr.zeros_like(xr_data.pv_system_id, dtype=bool)
                system_mask.values[:] = np.random.uniform(size=n_systems) >= dropout_p

                xr_data = xr_data.where(system_mask)

            if self.system_dropout_timedeltas is not None:
                # Apply independent delay to each PV system
                delay_mask = xr.zeros_like(xr_data, dtype=bool)

                last_available_times = xr.zeros_like(
                    xr_data.pv_system_id, dtype=xr_data.time_utc.dtype
                )
                last_available_times.values[:] = t0 + np.random.choice(
                    self.system_dropout_timedeltas, size=n_systems
                )

                delay_mask = xr_data.time_utc <= last_available_times

                # Apply masking
                xr_data = xr_data.where(delay_mask)

            yield xr_data



"""Ensure there is N PV systems per example"""

import logging

import numpy as np
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("ensure_n_gsp_per_example")
class EnsureNGSPSPerExampleIterDataPipe(IterDataPipe):
    """Ensure there is N GSP per example"""

    def __init__(self, source_datapipe: IterDataPipe, n_gsps_per_example: int, seed=None):
        """
        Ensure there is N GSPS per example

        Args:
            source_datapipe: Datapipe of GSP data
            n_gsps_per_example: Number of GSPS to have in example
            seed: Random seed for choosing
        """
        self.source_datapipe = source_datapipe
        self.n_gsps_per_example = n_gsps_per_example
        self.rng = np.random.default_rng(seed=seed)

    def __iter__(self):
        for xr_data in self.source_datapipe:
            if len(xr_data.gsp_id) > self.n_gsps_per_example:
                logger.debug(f"Reducing GSPS to  {self.n_gsps_per_example}")
                # More PV systems are available than we need. Reduce by randomly sampling:
                subset_of_gsp_ids = self.rng.choice(
                    xr_data.gsp_id,
                    size=self.n_gsps_per_example,
                    replace=False,
                )
                xr_data = xr_data.sel(gsp_id=subset_of_gsp_ids)
            elif len(xr_data.gsp_id) < self.n_gsps_per_example:
                logger.debug("Padding out GSP")
                # If we just used `choice(replace=True)` then there's a high chance
                # that the output won't include every available PV system but instead
                # will repeat some PV systems at the expense of leaving some on the table.
                # TODO: Don't repeat GSP. Pad with NaNs and mask the loss. Issue #73.
                assert len(xr_data.gsp_id) > 0, (
                    "There are no GSPS at all. " "We need at least one in an example"
                )
                n_random_gsps = self.n_gsps_per_example - len(xr_data.gsp_id)
                allow_replacement = n_random_gsps > len(xr_data.gsp_id)
                random_gsp_ids = self.rng.choice(
                    xr_data.gsp_id,
                    size=n_random_gsps,
                    replace=allow_replacement,
                )
                xr_data = xr.concat(
                    (xr_data, xr_data.sel(gsp_id=random_gsp_ids)),
                    dim="gsp_id",
                )
            yield xr_data



"""Numpy transforms"""

from .add_fourier_space_time import AddFourierSpaceTimeIterDataPipe as AddFourierSpaceTime
from .add_topographic_data import AddTopographicDataIterDataPipe as AddTopographicData
from .datetime_features import AddTrigonometricDateTimeIterDataPipe as AddTrigonometricDateTime
from .sun_position import AddSunPositionIterDataPipe as AddSunPosition



"""Conversion from Xarray to NumpyBatch"""

from .gsp import convert_gsp_to_numpy_batch
from .nwp import convert_nwp_to_numpy_batch
from .pv import convert_pv_to_numpy_batch
from .satellite import convert_satellite_to_numpy_batch
from .sensor import convert_sensor_to_numpy_batch
from .wind import convert_wind_to_numpy_batch



"""Datapipes from TorchData that have been copied in for use with core PyTorch Datapipes"""

from typing import Iterator, List, Optional, Sequence, Sized, Tuple, TypeVar

from torch.utils.data import IterDataPipe, functional_datapipe
from torch.utils.data.datapipes.iter.combining import T_co, _ChildDataPipe, _ForkerIterDataPipe

T = TypeVar("T")


# https://github.com/pytorch/data/issues/865
@functional_datapipe("zip_ocf")
class ZipperIterDataPipe(IterDataPipe[Tuple[T_co]]):
    """
    Aggregates elements into a tuple from each of the input DataPipes (functional name: ``zip``).

    The output is stopped as soon as the shortest input DataPipe is exhausted.

    Args:
        *datapipes: Iterable DataPipes being aggregated
    Example:
        >>> # xdoctest: +REQUIRES(module:torchdata)
        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> dp1, dp2, dp3 = IterableWrapper(range(5)), IterableWrapper(range(10, 15)),
        >>> IterableWrapper(range(20, 25))
        >>> list(dp1.zip(dp2, dp3))
        [(0, 10, 20), (1, 11, 21), (2, 12, 22), (3, 13, 23), (4, 14, 24)]
    """

    datapipes: Tuple[IterDataPipe]
    length: Optional[int]

    def __init__(self, *datapipes: IterDataPipe):
        """Init"""
        if not all(isinstance(dp, IterDataPipe) for dp in datapipes):
            raise TypeError(
                "All inputs are required to be `IterDataPipe` " "for `ZipIterDataPipe`."
            )
        super().__init__()
        self.datapipes = datapipes  # type: ignore[assignment]
        self.length = None

    def __iter__(self) -> Iterator[Tuple[T_co]]:
        """Iter"""
        iterators = [iter(datapipe) for datapipe in self.datapipes]
        for data in zip(*iterators):
            yield data

    def __len__(self) -> int:
        """Len"""
        if self.length is not None:
            if self.length == -1:
                raise TypeError("{} instance doesn't have valid length".format(type(self).__name__))
            return self.length
        if all(isinstance(dp, Sized) for dp in self.datapipes):
            self.length = min(len(dp) for dp in self.datapipes)
        else:
            self.length = -1
        return len(self)


@functional_datapipe("repeat")
class RepeaterIterDataPipe(IterDataPipe[T_co]):
    """
    Repeater

    Repeatedly yield each element of source DataPipe for the specified number of times before
    moving onto the next element (functional name: ``repeat``). Note that no copy is made
    in this DataPipe,
    the same element is yielded repeatedly.

    If you would like to yield the whole DataPipe in order multiple times, use :class:`.Cycler`.

    Args:
        source_datapipe: source DataPipe that will be iterated through
        times: the number of times an element of ``source_datapipe`` will be yielded
        before moving onto the next element

    Example:
        >>> from torchdata.datapipes.iter import IterableWrapper
        >>> dp = IterableWrapper(range(3))
        >>> dp = dp.repeat(2)
        >>> list(dp)
        [0, 0, 1, 1, 2, 2]
    """

    def __init__(self, source_datapipe: IterDataPipe[T_co], times: int) -> None:
        """Init"""
        self.source_datapipe: IterDataPipe[T_co] = source_datapipe
        self.times: int = times
        if times <= 1:
            raise ValueError(f"The number of repetition must be > 1, got {times}")

    def __iter__(self) -> Iterator[T_co]:
        """Iter"""
        for element in self.source_datapipe:
            for _ in range(self.times):
                yield element

    def __len__(self) -> int:
        """Len"""
        return self.times * len(self.source_datapipe)


@functional_datapipe("unzip")
class UnZipperIterDataPipe(IterDataPipe[T]):
    r"""
    UnZipper Iterator DataPipe

    Takes in a DataPipe of Sequences, unpacks each Sequence, and return the
    elements in separate DataPipes
    based on their position in the Sequence (functional name: ``unzip``).
    The number of instances produced equals to
    the sequence length minus the number of columns to skip.

    Note:
        Each sequence within the DataPipe should have the same length, specified by
        the input argument `sequence_length`.

    Args:
        source_datapipe: Iterable DataPipe with sequences of data
        sequence_length: Length of the sequence within the source_datapipe.
            All elements should have the same length.
        buffer_size: this restricts how far ahead the leading child DataPipe can read relative
            to the slowest child DataPipe. Use -1 for the unlimited buffer.
        columns_to_skip: optional indices of columns that the DataPipe should skip
            (each index should be an integer from 0 to sequence_length - 1)

    Example:
        >>> from torchdata.datapipes.iter import IterableWrapper
        >>> source_dp = IterableWrapper([(i, i + 10, i + 20) for i in range(3)])
        >>> dp1, dp2, dp3 = source_dp.unzip(sequence_length=3)
        >>> list(dp1)
        [0, 1, 2]
        >>> list(dp2)
        [10, 11, 12]
        >>> list(dp3)
        [20, 21, 22]
    """

    def __new__(
        cls,
        source_datapipe: IterDataPipe[Sequence[T]],
        sequence_length: int,
        buffer_size: int = 1000,
        columns_to_skip: Optional[Sequence[int]] = None,
    ):
        """Create a new instance"""
        if columns_to_skip is None:
            instance_ids = list(range(sequence_length))
        else:
            skips = set(columns_to_skip)
            instance_ids = [i for i in range(sequence_length) if i not in skips]

        if len(instance_ids) == 0:
            raise RuntimeError(
                "All instances are being filtered out in UnZipperIterDataPipe. Please check"
                "the input `sequence_length` and `columns_to_skip`."
            )

        # The implementation basically uses Forker but only yields a
        # specific element within the sequence
        container = _UnZipperIterDataPipe(source_datapipe, instance_ids, buffer_size)  # type: ignore[arg-type]
        return [_ChildDataPipe(container, i) for i in range(len(instance_ids))]


class _UnZipperIterDataPipe(_ForkerIterDataPipe):
    """Internal UnZipper"""

    def __init__(self, datapipe: IterDataPipe, instance_ids: List[int], buffer_size: int = 1000):
        """Init"""
        super().__init__(datapipe, len(instance_ids), buffer_size)  # type: ignore[arg-type]
        self.instance_ids = instance_ids

    def get_next_element_by_instance(self, instance_id: int):
        r"""
        Get next element by instance

        Note:
            Each element returned from the source datapipe is required to be a sequnce that can
            be subscribed with a column index
        """
        for return_val in super().get_next_element_by_instance(instance_id):
            yield return_val[self.instance_ids[instance_id]]

    def __getstate__(self):
        """Get state"""
        state = super().__getstate__()
        return (*state, self.instance_ids)

    def __setstate__(self, state):
        """Set state"""
        super().__setstate__(state[:-1])
        self.instance_ids = state[-1]


@functional_datapipe("set_length")
class LengthSetterIterDataPipe(IterDataPipe[T_co]):
    r"""
    Length setter

    Set the length attribute of the DataPipe, which is returned by
    ``__len__`` (functional name: ``set_length``).
    This can be used after DataPipes whose final length cannot be known in advance
    (e.g. ``filter``). If you
    know the final length with certainty, you can manually set it, which can then be used by
    DataLoader or other DataPipes.

    Note:
        This DataPipe differs from :class:`.Header` in that this doesn't restrict
        the number of elements that
        can be yielded from the DataPipe; this is strictly used for setting an attribute
        so that it can be used later.

    Args:
        source_datapipe: a DataPipe
        length: the integer value that will be set as the length

    Example:
        >>> from torchdata.datapipes.iter import IterableWrapper
        >>> dp = IterableWrapper(range(10)).filter(lambda x: x < 5).set_length(3)
        >>> list(dp)  # Notice that the number of elements yielded is unchanged
        [0, 1, 2, 3, 4]
        >>> len(dp)
        3
        >>> header_dp = IterableWrapper(range(10)).filter(lambda x: x < 5).header(3)
        >>> list(header_dp)  # Use `.header()` if you want to limit the number of elements yielded
        [0, 1, 2]
        >>> len(header_dp)
        3
    """

    def __init__(self, source_datapipe: IterDataPipe[T_co], length: int) -> None:
        """Init"""
        self.source_datapipe: IterDataPipe[T_co] = source_datapipe
        assert length >= 0
        self.length: int = length

    def __iter__(self) -> Iterator[T_co]:
        """Iter"""
        yield from self.source_datapipe

    def __len__(self) -> int:
        """Len"""
        return self.length


@functional_datapipe("header")
class HeaderIterDataPipe(IterDataPipe[T_co]):
    r"""
    Header Iterator DataPipe

    Yields elements from the source DataPipe from the start, up to the specfied
    limit (functional name: ``header``).

    If you would like to manually set the length of a DataPipe to a certain value;
     we recommend you to
    use :class:`.LengthSetter`.

    Args:
        source_datapipe: the DataPipe from which elements will be yielded
        limit: the number of elements to yield before stopping

    Example:
        >>> from torchdata.datapipes.iter import IterableWrapper
        >>> dp = IterableWrapper(range(10))
        >>> header_dp = dp.header(3)
        >>> list(header_dp)
        [0, 1, 2]
    """

    def __init__(self, source_datapipe: IterDataPipe[T_co], limit: Optional[int] = 10) -> None:
        """Init"""
        self.source_datapipe: IterDataPipe[T_co] = source_datapipe
        self.limit: Optional[int] = limit

    def __iter__(self) -> Iterator[T_co]:
        """Iter"""
        i: int = 0
        for value in self.source_datapipe:
            i += 1
            if self.limit is None or i <= self.limit:
                yield value
            else:
                break

    def __len__(self) -> int:
        """Len"""
        try:
            source_len = len(self.source_datapipe)
            return source_len if self.limit is None else min(source_len, self.limit)
        except TypeError as error:
            if self.limit is None:
                raise TypeError(
                    "The length of this HeaderIterDataPipe cannot be determined."
                ) from error
            return self.limit



"""Datapipes to check equalities"""

import logging
from typing import Optional, Union

import numpy as np
import xarray as xr
from torch.utils.data import IterDataPipe, functional_datapipe

logger = logging.getLogger(__name__)


@functional_datapipe("check_value_equal_to_fraction")
class CheckValueEqualToFractionIterDataPipe(IterDataPipe):
    """Check how much of the input is equal to a given fraction"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        value: int,
        fraction: float,
        dataset_name: Optional[str] = None,
    ):
        """
        Check how much of the input is equal to a given fraction

        Args:
            source_datapipe: Datapipe emitting Xarray object
            value: Value to check
            fraction: Fraction for threshold
            dataset_name: Optional dataset name if checking a subset
        """
        self.source_datapipe = source_datapipe
        self.value = value
        self.fraction = fraction
        self.dataset_name = dataset_name

    def __iter__(self) -> Union[xr.Dataset, xr.DataArray]:
        """Check equality"""
        for xr_data in self.source_datapipe:
            self.check_fraction_of_dataset_equals_value(
                xr_data[self.dataset_name] if self.dataset_name is not None else xr_data,
                value=self.value,
                fraction=self.fraction,
            )
            yield xr_data

    def check_fraction_of_dataset_equals_value(self, data: xr.Dataset, value: int, fraction: float):
        """Check data is greater than a certain value"""
        # Get elementwise equality and fraction
        data_equal_to_value = np.isclose(data, value)
        fraction_equal_to_value = data_equal_to_value.mean()
        # Check fraction is greater than threshold
        if fraction_equal_to_value > fraction:
            message = (
                f"Fraction of data equal to {value} is greater than {fraction} in "
                f"{self.source_datapipe.__repr__()} "
            )
            if self.dataset_name is not None:
                message += f"dataset {self.dataset_name}. "
            message += f"The fraction is {fraction_equal_to_value}. "
            raise Warning(message)


@functional_datapipe("check_greater_than_or_equal_to")
class CheckGreaterThanOrEqualToIterDataPipe(IterDataPipe):
    """Check greater than or equal to"""

    def __init__(
        self, source_datapipe: IterDataPipe, min_value: int, dataset_name: Optional[str] = None
    ):
        """
        Check greater than or equal to check

        Args:
            source_datapipe: Datapipe emitting Xarray object
            min_value: Minimum value
            dataset_name: Optional dataset name if checking a subset
        """
        self.source_datapipe = source_datapipe
        self.min_value = min_value
        self.dataset_name = dataset_name

    def __iter__(self) -> Union[xr.Dataset, xr.DataArray]:
        """Check equality"""
        for xr_data in self.source_datapipe:
            self.check_dataset_greater_than_or_equal_to(
                xr_data[self.dataset_name] if self.dataset_name is not None else xr_data,
                min_value=self.min_value,
            )
            yield xr_data

    def check_dataset_greater_than_or_equal_to(self, data: xr.Dataset, min_value: int):
        """Check data is greater than a certain value"""
        if (data < min_value).any():
            message = (
                f"Some data values are less than {min_value} in {self.source_datapipe.__repr__()} "
            )
            if self.dataset_name is not None:
                message += f"dataset {self.dataset_name}. "
            message += f"The minimum value is {data.min()}. "
            raise Warning(message)


@functional_datapipe("check_less_than_or_equal_to")
class CheckLessThanOrEqualToIterDataPipe(IterDataPipe):
    """Check less than or equal to equality"""

    def __init__(
        self, source_datapipe: IterDataPipe, max_value: int, dataset_name: Optional[str] = None
    ):
        """
        Check less than or equal to equality

        Args:
            source_datapipe: Datapipe emitting Xarray object
            max_value: Max value
            dataset_name: Optioanl dataset name if checking a subset
        """
        self.source_datapipe = source_datapipe
        self.max_value = max_value
        self.dataset_name = dataset_name

    def __iter__(self) -> Union[xr.Dataset, xr.DataArray]:
        """Check equality"""
        for xr_data in self.source_datapipe:
            self.check_dataset_less_than_or_equal_to(
                xr_data[self.dataset_name] if self.dataset_name is not None else xr_data,
                max_value=self.max_value,
            )
            yield xr_data

    def check_dataset_less_than_or_equal_to(self, data: xr.Dataset, max_value: int):
        """Check data is less than a certain value"""
        if (data > max_value).any():
            message = (
                f"Some data values are more than {max_value} in {self.source_datapipe.__repr__()} "
            )
            if self.dataset_name is not None:
                message += f"dataset {self.dataset_name}. "
            message += f"The maximum value  is {data.max()}. "
            raise Warning(message)


@functional_datapipe("check_not_equal_to")
class CheckNotEqualToIterDataPipe(IterDataPipe):
    """Check not equal to equality"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        value: int,
        dataset_name: Optional[str] = None,
        raise_error: bool = False,
    ):
        """
        Checks not equal to equality on the data

        Args:
            source_datapipe: Datapipe emitting Xarray object
            value: Value to check
            dataset_name: Optional dataset name if checkinga subset
            raise_error: Whether to raise an error or not
        """
        self.source_datapipe = source_datapipe
        self.value = value
        self.dataset_name = dataset_name
        self.raise_error = raise_error

    def __iter__(self) -> Union[xr.Dataset, xr.DataArray]:
        """Check not equal equality"""
        for xr_data in self.source_datapipe:
            self.check_dataset_not_equal(
                xr_data[self.dataset_name] if self.dataset_name is not None else xr_data,
                value=self.value,
                raise_error=self.raise_error,
            )
            yield xr_data

    def check_dataset_not_equal(self, data: xr.Dataset, value: int, raise_error: bool = True):
        """Check data is not equal than a certain value"""
        if np.isclose(data, value).any():
            message = f"Some data values are equal to {value} in {self.source_datapipe.__repr__()} "
            if self.dataset_name is not None:
                message += f"dataset {self.dataset_name}. "
            if raise_error:
                logger.error(message)
                raise Exception(message)
            else:
                logger.warning(message)
